"""Constant distributions and fitters."""
import datetime as dt

import numpy as np
import polars as pl

from metasyn.distribution.base import (
    BaseDistribution,
    BaseFitter,
    builtin_fitter,
    convert_to_series,
    metadist,
)
from metasyn.distribution.util import convert_numpy_datetime


class BaseConstantDistribution(BaseDistribution):
    """Base class for constant distribution.

    This base class makes it easy to implement new constant distributions
    for different variable types.
    """

    def __init__(self, value) -> None:
        self.value = value

    def _param_dict(self):
        return {"value": self.value}

    def draw(self):
        return self.value

    def information_criterion(self, values):
        vals = convert_to_series(values)
        return -np.inf if vals.n_unique() < 2 else np.inf


class BaseConstantFitter(BaseFitter):
    """Base distribution for many constant fitters."""

    distribution: type[BaseConstantDistribution]

    def _fit(self, values: pl.Series) -> BaseDistribution:
        # if unique, just get that value
        if values.n_unique() == 1:
            return self.distribution(values.unique()[0])

        # otherwise get most common value
        val = values.value_counts(sort=True)[0,0]
        return self.distribution(val)


@metadist(name="core.constant", var_type="continuous")
class ContinuousConstantDistribution(BaseConstantDistribution):
    """Constant distribution for floating point type.

    This class implements the constant distribution, so that it draws always
    the same value.

    Parameters
    ----------
    value: float
        Value that will be returned when drawn.

    Examples
    --------
    >>> ConstantDistribution(2.45)
    """

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution:  # noqa: ARG003
        return cls(0.0)

    @classmethod
    def _param_schema(cls):
        return {
            "value": {"type": "number"}
        }


@metadist(name="core.constant", var_type="discrete")
class DiscreteConstantDistribution(BaseConstantDistribution):
    """Constant discrete distribution.

    This class implements the constant distribution, so that it draws always
    the same value.

    Parameters
    ----------
    value: int
        Value that will be returned when drawn.

    Examples
    --------
    >>> DiscreteConstantDistribution(213456)
    """

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls(0)

    @classmethod
    def _param_schema(cls):
        return {
            "value": {"type": "integer"}
        }


@metadist(name="core.constant", var_type="datetime")
class DateTimeConstantDistribution(BaseConstantDistribution):
    """Constant datetime distribution.

    This class implements the constant distribution, so that it draws always
    the same value.

    Parameters
    ----------
    value: str or datetime.datetime
        Value that will be returned when drawn.

    Examples
    --------
    >>> DateTimeConstantDistribution(value="2022-07-15T10:39:36")
    """

    def __init__(self, value):
        if isinstance(value, str):
            value = dt.datetime.fromisoformat(value)
        elif isinstance(value, np.datetime64):
            value = convert_numpy_datetime(value)
        super().__init__(value)

    def _param_dict(self):
        return {"value": self.value.isoformat()}

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls("2022-07-15T10:39:36")

    @classmethod
    def _param_schema(cls):
        return {
            "value": {"type": "string"}
        }


@metadist(name="core.constant", var_type="time")
class TimeConstantDistribution(BaseConstantDistribution):
    """Constant time distribution.

    This class implements the constant distribution, so that it draws always
    the same value.

    Parameters
    ----------
    value: str or datetime.time
        Value that will be returned when drawn.

    Examples
    --------
    >>> TimeConstantDistribution(value="10:39:36")
    """

    def __init__(self, value):
        if isinstance(value, str):
            value = dt.time.fromisoformat(value)
        super().__init__(value)

    def _param_dict(self):
        return {"value": self.value.isoformat()}

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls("10:39:36")

    @classmethod
    def _param_schema(cls):
        return {
            "value": {"type": "string"}
        }


@metadist(name="core.constant", var_type="date")
class DateConstantDistribution(BaseConstantDistribution):
    """Constant date distribution.

    This class implements the constant distribution, so that it draws always
    the same value.

    Parameters
    ----------
    value: str or datetime.date
        Value that will be returned when drawn.

    Examples
    --------
    >>> DateConstantDistribution(value="1903-07-15")
    """

    def __init__(self, value):
        if isinstance(value, str):
            value = dt.date.fromisoformat(value)
        super().__init__(value)

    def _param_dict(self):
        return {"value": self.value.isoformat()}

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls("1903-07-15")

    @classmethod
    def _param_schema(cls):
        return {
            "value": {"type": "string"}
        }


@metadist(name="core.constant", var_type="string")
class StringConstantDistribution(BaseConstantDistribution):
    """Constant string distribution.

    This class implements the constant distribution, so that it draws always
    the same value.

    Parameters
    ----------
    value: str
        Value that will be returned when drawn.

    Examples
    --------
    >>> ConstantDistribution("some_string")

    """

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls("text")

    @classmethod
    def _param_schema(cls):
        return {
            "value": {"type": "string"}
        }


@builtin_fitter(distribution=DiscreteConstantDistribution, var_type="discrete")
class DiscreteConstantFitter(BaseConstantFitter):
    """Fitter for constant discrete distribution."""

@builtin_fitter(distribution=ContinuousConstantDistribution, var_type="continuous")
class ContinuousConstantFitter(BaseConstantFitter):
    """Fitter for constant continuous distribution."""

@builtin_fitter(distribution=DateConstantDistribution, var_type="date")
class DateConstantFitter(BaseConstantFitter):
    """Fitter for constant date distribution."""

@builtin_fitter(distribution=TimeConstantDistribution, var_type="time")
class TimeConstantFitter(BaseConstantFitter):
    """Fitter for constant time distribution."""

@builtin_fitter(distribution=DateTimeConstantDistribution, var_type="datetime")
class DateTimeConstantFitter(BaseConstantFitter):
    """Fitter for constant datetime distribution."""

@builtin_fitter(distribution=StringConstantDistribution, var_type="string")
class StringConstantFitter(BaseConstantFitter):
    """Fitter for constant string distribution."""



"""File interfaces to read dataset and write synthetic datasets."""
from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional, Type, Union

import polars as pl

_AVAILABLE_FILE_INTERFACES = {}


def fileinterface(*args):
    """Register a dataset so that it can be found by name."""

    def _wrap(cls):
        _AVAILABLE_FILE_INTERFACES[cls.format] = cls
        return cls

    return _wrap(*args)

class BaseFileInterface(ABC):
    """Abstract file interface class to derive specific implementations from.

    The file interface facilitates the reading and writing of dataset files. In
    particular they can ensure that the output data are exactly the same as the
    input data.

    The implementation class should have at least two class attributes: a :code:`name`
    for the implementation and :code:`extensions`, which is a list of extensions to be
    associated with the implementation. For example :code:`[".csv", ".tsv"]`.
    """

    format = "base"
    extensions: list[str] = []

    def __init__(self, metadata: dict[str, Any], file_name: str):
        """Initialize the file interface with metadata and the original file name.

        Parameters
        ----------
        metadata:
            A dictionary containing all the information such as metadata and
            file format directives. The structure of the metadata is determined
            by the implementation of the BaseFileInterface and can be empty.
        file_name:
            file name of the original dataset.
        """
        self.metadata = metadata
        self.file_name = file_name

    def to_dict(self) -> dict[str, Any]:
        """Convert the class instance to a dictionary.

        Returns
        -------
            A dictionary containing all information to reconstruct the file interface.
        """
        if self.format not in _AVAILABLE_FILE_INTERFACES:
            warnings.warn(f"Current file interface {self.format} is not available, "
                          "did you forget to use"
                          f" the decorator @fileinterface for the class {self.__class__}?")
        if self.format == "base":
            warnings.warn(f"Class attribute for {self.__class__} should not be 'base'."
                           " Please give another name to your file interface.")
        return {
            "file_interface_name": self.format,
            "format_metadata": self.metadata,
            "file_name": self.file_name,
        }

    @abstractmethod
    def _write_file(self, df: pl.DataFrame, fp: Union[Path, str]):
        """Write synthetic file, to be implemented by file interface implementations.

        Parameters
        ----------
        df:
            Polars dataframe to be written to a file.
        fp:
            Name of the file to be written to.

        """
        raise NotImplementedError("Write synthetic is not implemented for the BaseFileInterface.")

    def write_file(self, df: pl.DataFrame, fp: Union[None, Path, str] = None,
                        overwrite: bool = False):
        """Write the synthetic dataframe to a file.

        Parameters
        ----------
        df
            Dataframe to be written to a file.
        fp:
            File to write the dataframe to, by default None in which case
            the file will be the same as the original filename in the current
            working directory.
        overwrite:
            Allow overwriting of the file if it already exists, by default False.

        Raises
        ------
        FileExistsError
            If the file already exists and the overwrite argument is False.
        """
        fp = self.check_filename(fp, overwrite=overwrite)
        self._write_file(df, fp)

    def check_filename(self,  fp: Union[None, Path, str] = None,
                        overwrite: bool = False) -> Union[Path, str]:
        """Check whether the filename can be written to.

        Parameters
        ----------
        fp
            File check the filename for, by default None
        overwrite
            Whether overwriting is allowed, by default False

        Returns
        -------
            filename which could be either the same or different from fp.

        Raises
        ------
        FileExistsError
            If the file already exists and overwrite=False
        FileNotFoundError:
            If the parent directory of fp does not exist.
        """
        if fp is None:
            fp = self.file_name
        if Path(fp).is_file() and not overwrite:
            raise FileExistsError(f"File '{fp}' already exists, choose a different name or write "
                                  "to a different directory.")
        elif Path(fp).is_dir():
            fp = Path(fp) / self.file_name
        elif not Path(fp).parent.is_dir():
            raise FileNotFoundError(f"Parent directory does not exist for '{fp}'.")
        return fp

    @classmethod
    @abstractmethod
    def default_interface(cls, fp: Union[Path, str]):
        """Create a defeault interface with the most likely settings for writing.

        Parameters
        ----------
        fp
            File for writing to by default.

        Returns
        -------
            An instantiated file interface with default settings.
        """
        raise NotImplementedError("Default_interface method is not implemented for the base class.")

    @classmethod
    @abstractmethod
    def read_file(cls, fp: Union[Path, str]):
        """Create a file interface from a path.

        Parameters
        ----------
        fp
            Path to read the dataset from.build

        Returns
        -------
            An initialized file interface.

        """
        raise NotImplementedError("read_file method is not implemented for the base class.")


class ReadStatInterface(BaseFileInterface, ABC):
    """Abstract class to make it easier to create pyreadstat file interfaces."""

    interface = "unknown"

    def read_dataset(self, fp: Union[Path, str]):
        """Read the dataset without the metadata."""
        df, _ = SavFileInterface._get_df_metadata(fp)
        return df

    @classmethod
    def _read_data(cls, fp, max_rows=None, chunk_size=None):
        try:
            import pyreadstat  # noqa: PLC0415
        except ImportError as err:
            raise ImportError(
                f"Please install pyreadstat to use the {'/'.join(cls.extensions)} file interface."
                ) from err

        prs_func = getattr(pyreadstat, f"read_{cls.interface}")
        if max_rows is None:  # Read everything
            return prs_func(fp, apply_value_formats=True, output_format="polars")
        if chunk_size is None:  # Read first max_rows rows
            return prs_func(fp, apply_value_formats=True, output_format="polars",
                            row_limit=max_rows)

        _, metadata = prs_func(fp, metadataonly=True)
        n_rows = metadata.number_rows
        if max_rows >= 2*n_rows:  # Not enough rows to used chunked sampling, read first max_rows
            return prs_func(fp, apply_value_formats=True, output_format="polars",
                            row_limit=max_rows)

        skip_factor = n_rows // max_rows
        all_df = []
        i_chunk = 0
        for temp_df, prs_meta in pyreadstat.read_file_in_chunks(
                prs_func, fp, apply_value_formats=True, output_format="polars",
                chunksize=chunk_size):
            # Done
            if (i_chunk//skip_factor)*chunk_size >= max_rows:
                break
            if i_chunk % skip_factor == 0:
                all_df.append(temp_df)
            i_chunk += 1

        return pl.concat(all_df, how="vertical_relaxed"), prs_meta

    @classmethod
    def _get_df_metadata(cls, fp: Union[Path, str], **kwargs):
        """Read the dataset including the metadata."""
        df, prs_metadata = cls._read_data(fp, **kwargs)
        # df = pl.DataFrame(pandas_df)
        return cls._convert_with_orig_format(df, prs_metadata), prs_metadata


    @classmethod
    def read_file(cls, fp: Union[Path, str], **kwargs):
        """Create the file interface from a .sav or .zsav file.

        Parameters
        ----------
        fp:
            File to read the dataframe and metadata from.

        Returns
        -------
        df:
            Polars dataframe with the converted columns.
        file_interface:
            An instance of the :class:`SavFileInterface` with the appropriate metadata.
        """
        if Path(fp).suffix not in cls.extensions:
            warnings.warn(f"Trying to read file '{fp}' with extension different from"
                          f" {'/'.join(cls.extensions)}")

        df, prs_metadata = cls._get_df_metadata(fp, **kwargs)
        return df, cls(cls._extract_metadata(prs_metadata, fp), Path(fp).name)

    def _write_file(self, df: pl.DataFrame, out_fp: Union[Path, str]):
        try:
            import pyreadstat  # noqa: PLC0415
        except ImportError as err:
            raise ImportError(
                f"Please install pyreadstat to write {'/'.join(self.extensions)} files.") from err
        pd_df = self._prep_df_for_writing(df)
        label_start = "This is a synthetic dataset created by metasyn. Original label: "
        metadata = {k: v for k, v in self.metadata.items()}
        orig_file_label = metadata.pop("file_label", None)
        orig_file_label = "" if orig_file_label is None else orig_file_label
        metadata["variable_format"] = self._get_format(pd_df, df)
        if not orig_file_label.startswith(label_start):
            file_label = label_start + orig_file_label
        else:
            file_label = orig_file_label
        getattr(pyreadstat, f"write_{self.interface}")(pd_df, out_fp, **metadata,
                                                       file_label=file_label)

    @classmethod
    def default_interface(cls, fp: Union[str, Path]):
        return cls({}, Path(fp).name)

    @classmethod
    @abstractmethod
    def _convert_with_orig_format(cls, df, prs_metadata):
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def _extract_metadata(cls, prs_metadata, fp):
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def _prep_df_for_writing(cls, df):
        raise NotImplementedError()

    def _get_format(self, pd_df, df):  # noqa: ARG002
        return self.metadata.get("variable_format", {})


@fileinterface
class SavFileInterface(ReadStatInterface):
    """File interface for .sav and .zsav files.

    Also stores the descriptions of the columns and makes sure that F.0 columns
    are converted to integers.
    """

    format = "sav"
    extensions = [".sav", ".zsav"]
    interface = "sav"

    @classmethod
    def _convert_with_orig_format(cls, df, prs_metadata):
        """Read the dataset including the metadata."""
        for col in df.columns:
            col_format = prs_metadata.original_variable_types[col]
            if (col_format.startswith("F") and col_format.endswith(".0")
                    and df[col].dtype == pl.Float64):
                df = df.with_columns(pl.col(col).cast(pl.Int64))
        return df

    @classmethod
    def _extract_metadata(cls, prs_metadata, fp):
        if Path(fp).suffix == ".zsav":
            compress = True
        else:
            compress = False
        metadata = {
            "column_labels": prs_metadata.column_labels,
            "variable_format": prs_metadata.original_variable_types,
            "compress": compress,
            "variable_display_width": prs_metadata.variable_display_width,
            "file_label": prs_metadata.file_label,
            "variable_value_labels": prs_metadata.variable_value_labels,
            "variable_measure": prs_metadata.variable_measure,
        }
        return metadata


    def _prep_df_for_writing(self, df):
        if "variable_format" in self.metadata:
            for col in df.columns:
                col_format = self.metadata["variable_format"][col]
                if (col_format.startswith("F") and not col_format.endswith(".0")
                        and df[col].dtype == pl.Float64):
                    n_round = int(col_format.split(".")[-1])
                    df = df.with_columns(pl.col(col).round(n_round))
        return df

    @classmethod
    def default_interface(cls, fp: Union[str, Path]):
        return cls({}, Path(fp).name)


@fileinterface
class StataFileInterface(ReadStatInterface):
    """File interface for .dta files."""

    format = "dta"
    extensions = [".dta"]
    interface = "dta"

    @classmethod
    def _convert_with_orig_format(cls, df, prs_metadata):
        for col in df.columns:
            col_format = prs_metadata.original_variable_types[col]
            if col_format == "%8.0g":
                df = df.with_columns(pl.col(col).cast(pl.Int32))
            elif col_format == "%12.0g":
                df = df.with_columns(pl.col(col).cast(pl.Int64))
            elif col_format == "%9.0g":
                df = df.with_columns(pl.col(col).cast(pl.Float32))
            elif col_format == "%10.0g":
                df = df.with_columns(pl.col(col).cast(pl.Float64))
            # elif col_format == "%td":
                # print(col, df[col])
        return df

    @classmethod
    def _extract_metadata(cls, prs_metadata, fp):  # noqa: ARG003
        metadata = {
            "column_labels": prs_metadata.column_labels,
            "variable_format": prs_metadata.original_variable_types,
            "file_label": prs_metadata.file_label,
            "variable_value_labels": prs_metadata.variable_value_labels,
        }
        return metadata

    def _prep_df_for_writing(self, df):
        return df


    def _get_format(self, pd_df, df):  # noqa: ARG002
        """Fill in the formats for which we don't know them in the metadata."""
        var_format = {}
        for col in pd_df.columns:
            if col in self.metadata.get("variable_format", {}):
                continue
            pd_dtype = str(pd_df[col].dtype)
            if str(pd_dtype).startswith(("Int", "UInt")):
                if not str(pd_dtype).endswith("64"):
                    var_format[col] = "%8.0g"
                else:
                    var_format[col] = "%12.0g"
            elif str(pd_dtype) == "Float32":
                var_format[col] = "%9.0g"
            elif str(pd_dtype) == "Float64":
                var_format[col] = "%10.0g"
            # The below doesn't work due to overflow problems.
            # elif df[col].dtype == pl.Date:
                # var_format[col] = "%td"
                # pd_df[col] = 0
        var_format.update(self.metadata.get("variable_format", {}))
        return var_format


@fileinterface
class CsvFileInterface(BaseFileInterface):
    """File interface to read and write CSV files."""

    format = "csv"
    extensions = [".csv", ".tsv"]

    def read_dataset(self, fp: Union[Path, str], **kwargs):
        """Read CSV file.

        Parameters
        ----------
        fp:
            File to be read with the file interface.
        kwargs:
            Extra keyword arguments to be passed to polars.

        """
        df = pl.read_csv(
            fp, try_parse_dates=True, infer_schema_length=10000,
            null_values=self.metadata["null_value"],
            ignore_errors=True,
            separator=self.metadata["separator"],
            quote_char=self.metadata["quote_char"],
            encoding=self.metadata.get("encoding", "utf-8"),
            **kwargs)
        return df

    @classmethod
    def read_file(cls, fp: Union[Path, str], separator: Optional[str] = None, eol_char: str = "\n",
                  quote_char: str = '"', null_values: Union[None, str, list[str]] = None,
                  encoding="utf-8", **kwargs):
        r"""Read a csv file.

        See :func:`read_csv` for more detail.

        Parameters
        ----------
        fp:
            File to be read.
        separator:
            Separator, by default None
        eol_char:
            End of line character, by default "\\n"
        quote_char:
            Quotation character, by default '"'
        null_values:
            Null values, by default None

        Returns
        -------
        df:
            Polars dataframe for the file.
        file_interface:
            File interface that read the dataset.
        """
        if Path(fp).suffix == ".tsv" and separator is None:
            separator = "\t"
        if separator is None:
            separator = ","
        if null_values is None:
            null_values = ["", "na", "NA", "N/A", "Na"]
        if isinstance(null_values, str):
            null_values = [null_values]

        pl_keywords = dict(
            try_parse_dates=True,
            infer_schema_length=10000,
            null_values=null_values,
            ignore_errors=True,
            separator=separator,
            quote_char=quote_char,
            eol_char=eol_char,
            encoding=encoding,
        )
        pl_keywords.update(kwargs)
        df = pl.read_csv(fp, **pl_keywords)  # type: ignore

        metadata = {
            "separator": separator,
            "line_terminator": eol_char,
            "quote_char": quote_char,
            "null_value": null_values[0],
            "encoding": encoding,
        }
        return df, cls(metadata, Path(fp).name)

    def _write_file(self, df, out_fp):
        meta_copy = {k: v for k, v in self.metadata.items()}
        encoding = meta_copy.pop("encoding", "utf-8")
        if encoding == "utf-8":
            df.write_csv(out_fp, **meta_copy)
        else:
            buffer = df.write_csv(**meta_copy)
            with open(out_fp, "w", encoding=encoding, errors="replace") as file_handle:
                file_handle.write(buffer)

    @classmethod
    def default_interface(cls, fp: Union[Path, str]):
        return cls({
            "separator": ",",
            "line_terminator": "\n",
            "quote_char": '"',
            "null_value": "",
        }, Path(fp).name)

@fileinterface
class ExcelFileInterface(BaseFileInterface):
    """File interface/writer for Microsoft Excel files."""

    format = "excel"
    extensions = [".xlsx", ".xls", ".xlsb"]

    @classmethod
    def read_file(cls, fp: Union[Path, str], sheet_name: Optional[str] = None):
        df = pl.read_excel(source=str(fp), sheet_name=sheet_name)
        return df, cls({"worksheet": sheet_name}, Path(fp).name)

    def _write_file(self, df, out_fp):
        df.write_excel(out_fp, **self.metadata)

    @classmethod
    def default_interface(cls, fp: Union[Path, str]):
        return cls({"worksheet": "Sheet1"}, Path(fp).name)


def file_interface_from_dict(file_format_dict: dict) -> BaseFileInterface:
    """Create a file interface from a dictionary.

    Parameters
    ----------
    file_format_dict:
        Dictionary containing information to create the file interface.
    """
    for interface_name, interface in _AVAILABLE_FILE_INTERFACES.items():
        if file_format_dict["file_interface_name"] == interface_name:
            return interface(metadata=file_format_dict["format_metadata"],
                             file_name=file_format_dict["file_name"])
    raise ValueError(f"Cannot find file interface with name '{interface_name}'.")

def read_file(fp: Union[Path, str], name: Optional[str] = None,
              arguments: Optional[dict] = None) -> tuple[pl.DataFrame, BaseFileInterface]:
    """Attempt to create file interface from a dataset.

    Default options will be used to read in the file.

    Parameters
    ----------
    fp
        Filename of the dataset to be read.

    Returns
    -------
    df:
        Dataframe with the dataset.
    file_interface:
        The file interface that has created the dataframe.

    Raises
    ------
    ValueError
        When the extension is unknown.
    """
    arguments = {} if arguments is None else arguments
    if name is None:
        return get_file_interface_class(fp).read_file(fp, **arguments)
    for interface_name, interface in _AVAILABLE_FILE_INTERFACES.items():
        if interface_name == name:
            return interface.read_file(fp, **arguments)
    raise ValueError(f"Cannot find file interface with name '{name}'.")


def get_file_interface_class(fp: Union[Path, str]) -> Type[BaseFileInterface]:
    """Get the file interface class from a filename."""
    suffix = Path(fp).suffix

    for interface_name, interface in _AVAILABLE_FILE_INTERFACES.items():
        if suffix in interface.extensions:
            return interface
    raise ValueError(f"Files with extension '{suffix}' are not supported.")

def read_csv(fp: Union[Path, str], separator: Optional[str] = None, eol_char: str = "\n",
             quote_char: str = '"', null_values: Union[str, list[str], None]=None,
             **kwargs) -> tuple[pl.DataFrame, CsvFileInterface]:
    r"""Create the file interface from a file.

    This function is a wrapper around
    `polars.read_csv <https://docs.pola.rs/api/python/dev/reference/api/polars.read_csv.html>`
    with different defaults for some of the keywords, but all keywords should be passed through.

    Parameters
    ----------
    fp:
        File to be read.
    separator
        Separator for the csv file, by default None in which case
        the separator will be a "," for .csv files and a "\\t" for
        .tsv files.
    eol_char:
        End of line character, by default "\\n"
    quote_char:
        Quotation character, by default '"'
    null_values:
        Values that will be replaced by nulls, by default None in which
        case the defaults of polars will be used ["", "na", "NA", "N/A", "Na"].
    kwargs:
        Extra keyword arguments to be passed through to polars.

    Returns
    -------
    df:
        Data frame read from the files.
    cls:
        CsvFileInterface instance containing information on how to write CSV files.

    """
    return CsvFileInterface.read_file(fp, separator=separator, eol_char=eol_char,
                                      quote_char=quote_char, null_values=null_values, **kwargs)


def _parse_fmt(file_format, fp, interface_class) -> BaseFileInterface:
    if file_format is None:
        file_format = interface_class.default_interface(fp)
    if isinstance(file_format, dict):
        file_format = interface_class(file_format, fp)
    return file_format

def write_csv(df: Union[pl.DataFrame],
              fp: Union[Path, str],
              file_format: Union[dict, BaseFileInterface, None] = None,
              overwrite: bool = False):
    """Write to a CSV file with the same file format.

    Parameters
    ----------
    df
        DataFrame to write to a file.
    fp
        File to write to.
    file_format
        File format to use for writing the file, by default None meaning to use the defaults.
    overwrite
        Whether to overwrite the file if it exists, by default False
    """
    file_format = _parse_fmt(file_format, fp, CsvFileInterface)
    file_format.write_file(df, fp, overwrite=overwrite)

def read_tsv(*args, **kwargs) -> tuple[pl.DataFrame, CsvFileInterface]:
    """Alias for :func:`read_csv`."""
    return read_csv(*args, **kwargs)


def write_tsv(*args, **kwargs):
    """Alias for :func:`write_csv`."""
    return write_csv(*args, **kwargs)


def read_sav(fp: Union[Path, str], max_rows: Optional[int] = None,
             chunk_size: Optional[int] = None) -> tuple[pl.DataFrame, SavFileInterface]:
    """Create the file interface from a .sav or .zsav file.

    Parameters
    ----------
    fp:
        File to read the dataframe and metadata from.
    max_rows:
        Maximum number of rows to read in.
    chunk_size:
        Perform row sampling with contiguous rows. Should be used in combination
        with the max_rows parameter, otherwise it is ignored.

    Returns
    -------
    df:
        Polars dataframe with the converted columns.
    """
    return SavFileInterface.read_file(fp, max_rows=max_rows, chunk_size=chunk_size)


def write_sav(df: Union[pl.DataFrame],
              fp: Union[Path, str],
              file_format: Union[dict, BaseFileInterface, None] = None,
              overwrite: bool = False):
    """Write to a SAV file with the same file format.

    Parameters
    ----------
    df
        DataFrame to write to a file.
    fp
        File to write to.
    file_format
        File format to use for writing the file, by default None meaning to use the defaults.
    overwrite
        Whether to overwrite the file if it exists, by default False
    """
    file_format = _parse_fmt(file_format, fp, SavFileInterface)
    file_format.write_file(df, fp, overwrite=overwrite)


def read_dta(fp: Union[Path, str], max_rows: Optional[int] = None,
             chunk_size: Optional[int] = None) -> tuple[pl.DataFrame, StataFileInterface]:
    """Read a .dta stata file into metadata and a DataFrame.

    Parameters
    ----------
    fp
        File to be read with .dta extension.
    max_rows:
        Maximum number of rows to read in.
    chunk_size:
        Perform row sampling with contiguous rows. Should be used in combination
        with the max_rows parameter, otherwise it is ignored.

    Returns
    -------
    df:
        Polars dataframe with the converted columns.
    """
    return StataFileInterface.read_file(fp, max_rows=max_rows, chunk_size=chunk_size)


def write_dta(df: Union[pl.DataFrame],
              fp: Union[Path, str],
              file_format: Union[dict, BaseFileInterface, None] = None,
              overwrite: bool = False):
    """Write to a DTA file with the same file format.

    Parameters
    ----------
    df
        DataFrame to write to a file.
    fp
        File to write to.
    file_format
        File format to use for writing the file, by default None meaning to use the defaults.
    overwrite
        Whether to overwrite the file if it exists, by default False
    """
    file_format = _parse_fmt(file_format, fp, StataFileInterface)
    file_format.write_file(df, fp, overwrite=overwrite)


def read_excel(fp: Union[Path, str]) -> tuple[pl.DataFrame, ExcelFileInterface]:
    """Read an excel file and create a file interface from that.

    Parameters
    ----------
    fp
        Excel file to read.

    Returns
    -------
    df:
        Polars dataframe representing the excel dataset.
    file_interface:
        An instance of the :class:`ExcelFileInterface` used for writing excel files.
    """
    return ExcelFileInterface.read_file(fp)


def write_excel(df: Union[pl.DataFrame],
                fp: Union[Path, str],
                file_format: Union[dict, BaseFileInterface, None] = None,
                overwrite: bool = False):
    """Write to a Excel file with the same file format.

    Parameters
    ----------
    df
        DataFrame to write to a file.
    fp
        File to write to.
    file_format
        File format to use for writing the file, by default None meaning to use the defaults.
    overwrite
        Whether to overwrite the file if it exists, by default False
    """
    file_format = _parse_fmt(file_format, fp, ExcelFileInterface)
    file_format.write_file(df, fp, overwrite=overwrite)



"""Module implementing normal distributions."""

import numpy as np
from scipy.optimize import minimize
from scipy.stats import lognorm, norm, truncnorm
from scipy.stats._continuous_distns import FitDataError

from metasyn.distribution.base import (
    BaseDistribution,
    BaseFitter,
    ScipyDistribution,
    ScipyFitter,
    builtin_fitter,
    metadist,
)


@metadist(name="core.normal", var_type="continuous")
class ContinuousNormalDistribution(ScipyDistribution):
    """Normal distribution for floating point type.

    This class implements the normal/gaussian distribution and takes
    the average and standard deviation as initialization input.

    Parameters
    ----------
    mean: float
        Mean of the normal distribution.
    sd: float
        Standard deviation of the normal distribution.

    Examples
    --------
    >>> NormalDistribution(mean=1.3, sd=4.5)
    """

    scipy_class = norm

    def __init__(self, mean: float, sd: float):
        self.par = {"mean": mean, "sd": sd}
        self.dist = norm(loc=mean, scale=max(sd, 1e-8))

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls(0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "mean": {"type": "number"},
            "sd": {"type": "number"},
        }

@builtin_fitter(distribution=ContinuousNormalDistribution, var_type="continuous")
class ContinuousNormalFitter(ScipyFitter):
    """Fitter for continuous normal distribution."""

@metadist(name="core.lognormal", var_type="continuous")
class LogNormalDistribution(ScipyDistribution):
    """Log-normal distribution for floating point type.

    This class implements the log-normal mu and sigma as initialization input.

    Parameters
    ----------
    mean: float
        Controls the mean of the distribution.
    sd: float
        Controls the width of the distribution.

    Examples
    --------
    >>> LogNormalDistribution(mean=-2.0, sd=4.5)
    """

    scipy_class = lognorm

    def __init__(self, mean: float, sd: float):
        self.par = {"mean": mean, "sd": sd}
        self.dist = lognorm(s=max(sd, 1e-8), scale=np.exp(mean))

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls(0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "mean": {"type": "number"},
            "sd": {"type": "number"},
        }

@builtin_fitter(distribution=LogNormalDistribution, var_type="continuous")
class LogNormalFitter(BaseFitter):
    """Fitter for log normal distribution."""

    def _fit(self, series):
        try:
            sd, _, scale = self.distribution.scipy_class.fit(series, floc=0)
        except FitDataError:
            return self.distribution(0, 1)
        return self.distribution(np.log(scale), sd)


@metadist(name="core.truncated_normal", var_type="continuous")
class ContinuousTruncatedNormalDistribution(ScipyDistribution):
    """Truncated normal distribution for floating point type.

    Parameters
    ----------
    lower:
        Lower bound of the truncated normal distribution.
    upper:
        Upper bound of the truncated normal distribution.
    mean:
        Mean of the non-truncated normal distribution.
    sd:
        Standard deviation of the non-truncated normal distribution.

    Examples
    --------
    >>> TruncatedNormalDistribution(lower=1.0, upper=3.5, mean=2.3, sd=5)
    """

    scipy_class = truncnorm

    def __init__(self, lower: float, upper: float,
                 mean: float, sd: float):
        self.par = {"lower": lower, "upper": upper,
                    "mean": mean, "sd": sd}
        a, b = (lower-mean)/sd, (upper-mean)/sd
        self.dist = truncnorm(a=a, b=b, loc=mean, scale=max(sd, 1e-8))

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls(-1, 2, 0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "number"},
            "upper": {"type": "number"},
            "mean": {"type": "number"},
            "sd": {"type": "number"},
        }

@builtin_fitter(distribution=ContinuousTruncatedNormalDistribution, var_type="continuous")
class ContinuousTruncatedNormalFitter(BaseFitter):
    """Fitter for continuous truncated normal fitter."""

    def _fit(self, series):
        lower = series.min() - 1e-8
        upper = series.max() + 1e-8
        return self._fit_with_bounds(series, lower, upper)

    def _fit_with_bounds(self, values, lower, upper):
        def minimizer(param):
            mean, sd = param
            a, b = (lower-mean)/sd, (upper-mean)/sd
            dist = truncnorm(a=a, b=b, loc=mean, scale=sd)
            return -np.sum(dist.logpdf(values))

        x_start = [(lower+upper)/2, (upper-lower)/4]
        mean, sd = minimize(minimizer, x_start,
                             bounds=[(None, None),
                                     ((upper-lower)/100, None)]).x
        return self.distribution(lower, upper, mean, sd)


@metadist(name="core.normal", var_type="discrete")
class DiscreteNormalDistribution(ContinuousNormalDistribution):
    """Normal discrete distribution.

    This class implements the normal/gaussian distribution and takes
    the average and standard deviation as initialization input.

    Parameters
    ----------
    mean: float
        Mean of the normal distribution.

    sd: float
        Standard deviation of the normal distribution.

    Examples
    --------
    >>> DiscreteNormalDistribution(mean=2.4, sd=1.2)
    """

    def draw(self):
        return int(super().draw())


@builtin_fitter(distribution=DiscreteNormalDistribution, var_type="discrete")
class DiscreteNormalFitter(ScipyFitter):
    """Fitter for discrete normal distribution."""


@metadist(name="core.truncated_normal", var_type="discrete")
class DiscreteTruncatedNormalDistribution(ContinuousTruncatedNormalDistribution):
    """Truncated normal discrete distribution.

    Parameters
    ----------
    lower: float
        Lower bound of the truncated normal distribution.
    upper: float
        Upper bound of the truncated normal distribution.
    mean: float
        Mean of the non-truncated normal distribution.
    sd: float
        Standard deviation of the non-truncated normal distribution.

    Examples
    --------
    >>> DiscreteTruncatedNormalDistribution(lower=1.2, upper=4.5, mean=2.3, sd=4.5)
    """

    def draw(self):
        return int(super().draw())

@builtin_fitter(distribution=DiscreteTruncatedNormalDistribution, var_type="discrete")
class DiscreteTruncatedNormalFitter(ContinuousTruncatedNormalFitter):
    """Fitter for discrete truncated normal distribution."""



"""Module providing a Command Line Interface (CLI) for metasyn.

It provides functionality to generate GMF (.json) metadata files,
synthetic data from GMF files and creating json schemas for GMF files.
"""
import argparse
import json
import pathlib
import sys
from argparse import RawDescriptionHelpFormatter
from importlib.metadata import entry_points, version
from typing import Optional

from metasyn import MetaFrame
from metasyn.config import MetaConfig
from metasyn.file import file_interface_from_dict, get_file_interface_class, read_file
from metasyn.validation import create_schema

EXAMPLE_CREATE_META="metasyn create-meta your_dataset.csv -o your_gmf_file.json --config your_config.toml" # noqa: E501
EXAMPLE_CREATE_TOML="metasyn create-meta your_dataset.csv -o your_gmf_file.toml --config your_config.toml" # noqa: E501
EXAMPLE_SYNTHESIZE="metasyn synthesize your_gmf_file.json -o your_synthetic_file.csv"

MAIN_HELP_MESSAGE = f"""
Metasyn CLI version {version("metasyn")}

Usage: metasyn [subcommand] [options]

Available subcommands:
    create-meta:
        Create a intermediate metadata file (GMF/.json/.toml). This file can later be used to
        create a new synthetic dataset with the `synthesize` subcommand.
    synthesize:
        Create a synthetic dataset from the intermediate metadata file (GMF).
        To create a metadata file from your original dataset, use the `create-meta` subcommand.
    schema:
        Generate json schema from plugins.


To create a synthetic dataset from your original dataset you have to create a metadata file
and use this file to create a synthetic dataset.

WARNING: For the best results it is recommended to use the Python API. Things can and will go
wrong reading your dataset, and during the creation of metadata.

Example usage:

{EXAMPLE_CREATE_META}
{EXAMPLE_CREATE_TOML}
{EXAMPLE_SYNTHESIZE}


Program information:
    -v, --version - display CLI version and exit
    -h, --help    - display this help file and exit
"""

ENTRYPOINTS = ["create-meta", "synthesize", "schema"]


def main(input_args: Optional[list[str]] = None) -> None:
    """CLI pointing to different entrypoints."""
    # show help by default, else consume first argument
    if input_args is None:
        input_args = sys.argv[1:]
    else:
        input_args = [str(x) for x in input_args]
    if len(input_args) == 0:
        input_args = ["--help"]

    subcommand = input_args[0]
    input_args.pop(0)
    if subcommand in ["-h", "--help"]:
        print(MAIN_HELP_MESSAGE)
    elif subcommand in ["-v", "--version"]:
        print(f"Metasyn CLI version {version('metasyn')}")

    # find the subcommand in this module and run it!
    elif subcommand == "synthesize":
        synthesize(input_args)
    elif subcommand == "schema":
        schema(input_args)
    elif subcommand == "create-meta":
        create_metadata(input_args)
    else:
        print(f"Invalid subcommand ({subcommand}). For help see metasyn --help")
        sys.exit(1)


def create_metadata(input_args) -> None:
    """Program to create and save metadata from a DataFrame to a GMF file (.json/.toml)."""
    parser = argparse.ArgumentParser(
        prog="metasyn create-meta",
        description=f"""Create a Generative Metadata Format file from a CSV file.
This metadata file can then be used to create a synthetic dataset with the `synthesize` subcommand.

Examples:

{EXAMPLE_CREATE_META}
{EXAMPLE_CREATE_TOML}
""",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "input",
        help="input file; a CSV file that you want to synthesize later.",
        type=pathlib.Path,
        default=None,
        nargs="?",
    )
    parser.add_argument(
        "--output", "-o",
        help="Metadata GMF output file: .json/.toml. This file can be used to synthesize data.",
        type=pathlib.Path,
        default=None,
        required=False,
    )
    parser.add_argument(
        "--config",
        help="Configuration file (*.toml) to improve the quality and/or privacy of the metadata.",
        type=pathlib.Path,
        default=None,
    )

    args, _ = parser.parse_known_args(input_args)
    if args.config is not None:
        meta_config = MetaConfig.from_toml(args.config)
    else:
        meta_config = None

    if args.input is None:
        if meta_config is None:
            raise parser.error("Please supply either an input dataset or a configuration file.")
        meta_frame = MetaFrame.from_config(meta_config)
    else:
        if meta_config is not None and meta_config.file_config is not None:
            data_frame, file_handler = read_file(args.input, **meta_config.file_config)
        else:
            data_frame, file_handler = read_file(args.input)
        meta_frame = MetaFrame.fit_dataframe(data_frame, config=meta_config)
        meta_frame.file_format = file_handler.to_dict()
    meta_frame.save(args.output)


def synthesize(input_args) -> None:
    """Program to generate synthetic data."""
    parser = argparse.ArgumentParser(
        prog="metasyn synthesize",
        description=f"""Synthesize data from a Generative Metadata Format .json/.toml file.
To create the metadata file from your dataset, use the `create-meta` subcommand.

Example: {EXAMPLE_SYNTHESIZE}
""",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "input",
        help="input file; typically .json or .toml adhering to the Generative Metadata Format",
        type=pathlib.Path,
    )
    parser.add_argument(
        "--output", "-o",
        help="output file (.csv, .feather, .parquet, .pkl, or .xlsx)",
        type=pathlib.Path,
        required=False,
    )
    parser.add_argument(
        "-n", "--num_rows",
        help="number of rows to synthesize",
        default=None,
        type=int,
        required=False,
    )
    parser.add_argument(
        "-s", "--seed",
        help="Seed for the generation of synthetic data.",
        type=int,
        default=None,
        required=False,
    )
    parser.add_argument(
        "-p", "--preview",
        help="preview six-row synthesized data frame in console and exit",
        action="store_true",
    )

    # parse the args without the subcommand
    args, _ = parser.parse_known_args(input_args)

    if not args.preview and not args.output:
        parser.error("Output file is required if you are not using the preview option.")

    # Create the metaframe from the GMF file
    try:
        meta_frame = MetaFrame.load(args.input)
    except (json.JSONDecodeError, UnicodeDecodeError) as _err:
        parser.error(
            f"Unable to parse the file '{args.input}'.\n\n"
            "Expecting a GMF/.json/.toml file as input.\n"
            "Did you perhaps provide your dataset?\n"
            "If so, please first create the metadata with the `create-meta` sub command.\n"
            "Otherwise your GMF file might be corrupted, and you should recreate it.")
        return

    if args.preview:
        # only print six rows and exit
        print(meta_frame.synthesize(6, seed=args.seed))
        return


    # Store the dataframe to file
    if meta_frame.file_format is not None:
        file_interface = file_interface_from_dict(meta_frame.file_format)
        if args.output.suffix not in file_interface.extensions:
            file_interface = get_file_interface_class(args.output).default_interface(args.output)
        meta_frame.write_synthetic(args.output, n=args.num_rows, seed=args.seed,
                                   file_format=file_interface)
    else:
        file_interface = get_file_interface_class(args.output).default_interface(args.output)
        meta_frame.write_synthetic(args.output, n=args.num_rows, seed=args.seed,
                                   file_format=file_interface)


def schema(input_args) -> None:
    """Program to generate json schema from plugins."""
    parser = argparse.ArgumentParser(
        prog="metasyn schema",
        description="Create Generative Metadata Format schema and print to console.",
    )

    parser.add_argument(
        "plugins",
        help="Plugins to include in the generated schema (default builtin)",
        nargs="*"
    )

    parser.add_argument(
        "-l", "--list",
        help="display available plugins and quit",
        action="store_true",
    )

    parser.add_argument(
        "-o", "--output",
        help="File to write the schema to.",
        type=pathlib.Path,
    )

    # parse the args without the subcommand
    args = parser.parse_args(input_args)

    # deduplicated list of plugins for schema
    plugins_avail = {entry.name for entry in entry_points(group="metasyn.distribution_registry")}

    if args.list:
        for a in plugins_avail:
            print(a)
        return

    plugins = {"builtin", *args.plugins}
    if len(plugins - plugins_avail) > 0:
        notfound = ", ".join(plugins - plugins_avail)
        pl_avail = ", ".join(plugins_avail - {"builtin"})
        errmsg = (
            f"\n  Requested plugin(s) not found: {notfound}"
            f"\n  Available plugins: {pl_avail}"
        )
        parser.error(errmsg)
    jsonschema = create_schema(list(plugins))
    if args.output is None:
        print(json.dumps(jsonschema, indent=4))
    else:
        with open(args.output, "w") as handle:
            json.dump(jsonschema, handle, indent=4)


if __name__ == "__main__":
    main()



"""Module defining configuration classes for creating MetaFrames."""
from __future__ import annotations

import warnings
from pathlib import Path
from typing import Iterable, Optional, Union

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # type: ignore  # noqa

from metasyn.registry import DistributionRegistry
from metasyn.varspec import DistributionSpec, VarDefaults, VarSpec


class MetaConfig():
    """Configuration class for creating MetaFrames.

    This class is used to create, manipulate, and retrieve configurations for
    individual variables in a MetaFrame. It also provides methods for loading
    configurations from .toml files and converting them to dictionaries.

    Parameters
    ----------
    var_specs:
        List of configurations for individual variables. The order does not
        matter for variables that are found in the DataFrame, but in the case
        of variables that are data-free, the order is also the order of columns
        for the eventual synthesized dataframe. See the VarSpecAccess class on
        how the dictionary can be constructed.
    plugins:
        Plugins to use when fitting distributions to variables.
    privacy:
        Privacy method/level to use as a default setting for the privacy. Can be
        overridden in the var_spec for a particular column.
    n_rows:
        Number of rows for synthesization at a later stage. Can be unspecified by
        leaving the value at None.
    """

    def __init__(
            self,
            var_specs: Union[list[dict], list[VarSpec]],
            plugins: Union[DistributionRegistry, list[str], str, None],
            defaults: Optional[dict] = None,
            n_rows: Optional[int] = None,
            file_config: Optional[dict] = None,
            config_version: str = "1.2"):
        self.var_specs = [self._parse_var_spec(v) for v in var_specs]
        self.plugins = plugins  # type: ignore
        self.n_rows = n_rows
        defaults = {} if defaults is None else defaults
        self.defaults = VarDefaults(**defaults)
        self.config_version = config_version
        self.file_config = file_config

    @staticmethod
    def _parse_var_spec(var_spec) -> VarSpec:
        if isinstance(var_spec, VarSpec):
            return var_spec
        return VarSpec.from_dict(var_spec)

    @property
    def plugins(self) -> DistributionRegistry:
        """Return the plugin list to be used for creating the metaframe."""
        return self._plugins

    @plugins.setter
    def plugins(self, plugins):
        if not isinstance(plugins, DistributionRegistry):
            self._plugins = DistributionRegistry.parse(plugins)
        else:
            self._plugins = plugins

    def update_varspecs(self, new_var_specs: Union[list[dict], list[VarSpec]]):
        new_var_specs = [self._parse_var_spec(v) for v in new_var_specs]
        for cur_new_var_spec in new_var_specs:
            # Check if currently in varspecs and pop if it exists.
            for i_var, old_var_spec in enumerate(self.var_specs):
                if old_var_spec.name == cur_new_var_spec.name:
                    self.var_specs.pop(i_var)
                    break
            self.var_specs.append(cur_new_var_spec)

    @classmethod
    def from_toml(cls, config_fp: Union[str, Path]) -> MetaConfig:
        """Create a MetaConfig class from a .toml file.

        Parameters
        ----------
        config_fp:
            Path to the file containing the configuration.

        Returns
        -------
        meta_config:
            A fully initialized MetaConfig instance.
        """
        try:
            with open(config_fp, "rb") as handle:
                config_dict = tomllib.load(handle)
        except FileNotFoundError as fnf_error:
            raise FileNotFoundError(f"It appears '{config_fp}' is not a valid filepath."
                                    f" Please provide a path to a .toml file to load a MetaConfig"
                                    f" from.") from fnf_error
        except tomllib.TOMLDecodeError as value_error:
            if Path(config_fp).suffix != ".toml":
                raise ValueError(f"It appears '{Path(config_fp).name}' is a"
                                 f" '{Path(config_fp).suffix}' file."
                                 f" To load a MetaConfig, "
                                 f"provide the configuration as a .toml file.") from value_error
            raise value_error
        var_list = config_dict.pop("var", [])
        n_rows = config_dict.pop("n_rows", None)
        plugins = config_dict.pop("plugins", ["builtin"])
        defaults = config_dict.pop("defaults", None)
        privacy = config_dict.pop("privacy", None)
        file_config = config_dict.pop("file", None)
        config_version = config_dict.pop("config_version", "1.0")
        if config_version not in ["1.0", "1.1", "1.2"]:
            warnings.warn(f"Trying to read configuration file with version {config_version}, "
                          "this version of metasyn only supports 1.0, 1.1 and 1.2.")
        if privacy is not None:
            if defaults is not None:
                raise ValueError("Error parsing configuration file: cannot have both [privacy]"
                                 " and [defaults] tables.")
            defaults = {"privacy": privacy}
        if len(config_dict) > 0:
            raise ValueError(f"Error parsing configuration file '{config_fp}'."
                             f" Unknown keys detected: '{list(config_dict)}'")
        return cls(var_list, plugins, defaults, n_rows=n_rows,
                   file_config=file_config,
                   config_version=config_version)

    def to_dict(self) -> dict:
        """Convert the configuration to a dictionary.

        Returns
        -------
        config_dict:
            Configuration in dictionary form.
        """
        return {
            "config_version": self.config_version,
            "plugins": self.plugins,
            "n_rows": self.n_rows,
            "defaults": self.defaults,
            "var": self.var_specs
        }

    def get(self, name: str) -> VarSpecAccess:
        """Create a VarSpecAccess object pointing to a var with that name.

        If the variable does not exist, then a new variable config is created that
        has the default values.

        Parameters
        ----------
        name:
            Name of the variable configuration to retrieve.

        Returns
        -------
        var_spec:
            A variable config access object.
        """
        for var_spec in self.var_specs:
            if var_spec.name == name:
                return VarSpecAccess(var_spec, self)
        return VarSpecAccess(VarSpec(name=name), self)

    def iter_var(self, exclude: Optional[list[str]] = None) -> Iterable[VarSpecAccess]:
        """Iterate over all variables in the configuration.

        Parameters
        ----------
        exclude:
            Exclude variables with names in that list.

        Returns
        -------
        var_spec:
            VarSpecAccess class for each of the available variable configurations.
        """
        exclude = exclude if exclude is not None else []
        for var_spec in self.var_specs:
            if var_spec.name not in exclude:
                yield VarSpecAccess(var_spec, self)


class VarSpecAccess():
    """Access for variable configuration object.

    They take into account what the defaults are from the MetaConfig object.
    Otherwise they pass through all the attributes as normal and thus behave
    exactly as a variable config object themselves.

    Parameters
    ----------
    var_spec
        The variable configuration to access.
    meta_config
        The meta configuration instance to get default values from.
    """

    def __init__(self, var_spec: VarSpec, meta_config: MetaConfig):
        self.var_spec = var_spec
        self.meta_config = meta_config

    def __getattribute__(self, attr):
        if attr in ["privacy", "data_free", "prop_missing"]:
            if getattr(self.var_spec, attr) is None:
                return getattr(self.meta_config.defaults, attr)
            return getattr(self.var_spec, attr)
        if attr == "dist_spec":
            if self.data_free and self.var_spec.dist_spec.name is None:
                if self.var_type not in self.meta_config.defaults.distribution:
                    raise ValueError(
                        f"Variable with name '{self.name}' is declared datafree and a distribution "
                        f"cannot be inferred. You can set the distribution for '{self.name}' or set"
                        f" a default distribution for variable type '{self.var_type}'.")
                return DistributionSpec.parse(self.meta_config.defaults.distribution[self.var_type])
            return self.var_spec.dist_spec

        if attr not in ("var_spec", "meta_config") and hasattr(self.var_spec, attr):
            return getattr(self.var_spec, attr)
        return super().__getattribute__(attr)



from pathlib import Path

import polars as pl
import pytest
from pytest import mark

import metasyn as ms
from metasyn.demo.dataset import _AVAILABLE_DATASETS, demo_dataframe, demo_file
from metasyn.file import (
    _AVAILABLE_FILE_INTERFACES,
    BaseFileInterface,
    CsvFileInterface,
    SavFileInterface,
    file_interface_from_dict,
    fileinterface,
    read_csv,
    read_dta,
    read_file,
    read_sav,
    read_tsv,
)


class BadFileInterface1(BaseFileInterface):
    format = "bad1"
    def _write_file(self, df, fp):
        pass

    @classmethod
    def default_interface(cls, fp):
        return BaseFileInterface.default_interface(fp)

    @classmethod
    def read_file(cls, fp):
        return BaseFileInterface.read_file(fp)


@fileinterface
class BadFileInterface2(BaseFileInterface):
    def _write_file(self, df, fp):
        pass

    @classmethod
    def default_interface(cls, fp):
        pass

    @classmethod
    def read_file(cls, fp):
        pass

    @classmethod
    def default_interface(cls, fp):
        pass

    @classmethod
    def read_file(cls, fp):
        pass

@fileinterface
class BadFileInterface3(BaseFileInterface):
    format = "also_bad"
    def _write_file(self, df, fp):
        super()._write_file(df, fp)

    @classmethod
    def default_interface(cls, fp):
        pass

    @classmethod
    def read_file(cls, fp):
        pass

def test_decorator_warning():
    with pytest.warns(UserWarning):
        BadFileInterface1({}, "x").to_dict()
    with pytest.warns(UserWarning):
        BadFileInterface2({}, "x").to_dict()
        print(BadFileInterface2({}, "x").format)

def test_notimplemented():
    with pytest.raises(NotImplementedError):
        BadFileInterface3({}, "x").write_file(None)

def test_file_exists_error():
    with pytest.raises(FileExistsError):
        BadFileInterface1({}, "x").write_file(None, Path("tests", "data", "titanic.csv")
                                                , overwrite=False)

def test_dir_does_not_exist_error():
    with pytest.raises(FileNotFoundError):
        BadFileInterface1({}, "x").write_file(None, Path("tests", "data", "more", "directories",
                                                         "titanic.csv")
                                                , overwrite=False)

def test_default_interface_error():
    with pytest.raises(NotImplementedError):
        BadFileInterface1.default_interface("some_file")

def test_read_file_error():
    with pytest.raises(NotImplementedError):
        BadFileInterface1.read_file("some_file")

def test_sav_warning():
    with pytest.warns(UserWarning):
        SavFileInterface.read_file(Path("tests", "data", "actually_a_sav_file.csv"))


@mark.parametrize("filename",
                  ["SAQ (Item 3 Reversed).sav", "GlastonburyFestival.sav",
                   "Drug.sav", "butlerpsych.sav", "twowayexample.sav"])
def test_sav_interface(filename, tmpdir):
    sav_fp = Path("tests", "data", filename)
    df = SavFileInterface({},"x" ).read_dataset(sav_fp)
    assert isinstance(df, pl.DataFrame)
    df, sav_interface = read_sav(sav_fp)
    assert isinstance(df, pl.DataFrame)
    assert isinstance(sav_interface, SavFileInterface)
    assert len(df) > 0
    sav_interface.write_file(df, tmpdir)
    assert Path(tmpdir, filename).is_file()

    sav_interface.metadata["compress"] = True
    zsav_file = Path(tmpdir, Path(filename).stem + ".zsav")
    sav_interface.write_file(df, zsav_file)
    new_df, new_sav_interface = SavFileInterface.read_file(zsav_file)
    assert new_sav_interface.metadata["compress"]
    assert new_df.columns == df.columns


def test_prs_chunking():
    sav_fp = Path("tests", "data", "GlastonburyFestival.sav")
    df, _ = ms.read_sav(sav_fp)
    assert len(df) == 810
    df, _ = ms.read_sav(sav_fp, max_rows=10)
    assert len(df) == 10
    df, _ = ms.read_sav(sav_fp, max_rows=20, chunk_size=2)
    assert len(df) == 20

@mark.parametrize("dataset_name",
                  _AVAILABLE_DATASETS)
def test_csv_interface(dataset_name, tmpdir):
    filename = demo_file(dataset_name)
    direct_df, _ = CsvFileInterface.read_file(filename)
    assert isinstance(direct_df, pl.DataFrame)
    df, csv_interface = read_csv(filename)
    assert isinstance(df, pl.DataFrame)
    assert isinstance(csv_interface, CsvFileInterface)
    assert len(df) == len(direct_df) and len(df) > 0

    csv_interface.write_file(df, tmpdir)
    new_df, _ = CsvFileInterface.read_file(Path(tmpdir, Path(filename).name))
    assert isinstance(new_df, pl.DataFrame)
    assert df.columns == new_df.columns

    assert isinstance(csv_interface.read_dataset(filename), pl.DataFrame)

def test_tsv_interface(tmpdir):
    filename = Path("tests", "data", "data.tsv")

    df, tsv_interface = read_tsv(filename)
    assert tsv_interface.metadata["separator"] == "\t"
    tsv_interface.write_file(df, tmpdir)
    assert Path(tmpdir, "data.tsv").is_file()

def test_csv_null():
    filename = demo_file("hospital")
    df, csv_interface = read_csv(filename)
    assert len(df["age"].drop_nulls()) == 11
    df, csv_interface = read_csv(filename, null_values="84")
    assert len(df["age"].drop_nulls()) == 10

def test_csv_encoding(tmpdir):
    df, fmt = ms.read_csv(demo_file("test"), encoding="windows-1252")
    ms.write_csv(df, tmpdir / "temp.csv", file_format=fmt)

@mark.parametrize(
    "filename,interface_class",
    [(demo_file("hospital"), CsvFileInterface),
    (Path("tests", "data", "Drug.sav"), SavFileInterface),
    (Path("tests", "data", "data.tsv"), CsvFileInterface)]
)
def test_file_interface_from_dict(filename, interface_class):
    df, file_interface = read_file(filename)
    assert isinstance(file_interface, interface_class)
    assert isinstance(file_interface_from_dict(file_interface.to_dict()), interface_class)

def test_file_interface_errors():
    with pytest.raises(ValueError):
        file_interface_from_dict({"file_interface_name": "unknown"})

    with pytest.raises(ValueError):
        read_file(Path("tests", "test_fileinterface.py"))

@mark.parametrize("interface_class",
                  [x for x in _AVAILABLE_FILE_INTERFACES.values() if not x.__name__.startswith("Bad")])
def test_default_file_interfaces(interface_class, tmpdir):
    df = demo_dataframe("test")
    suffix = interface_class.extensions[0]
    fp = Path(tmpdir/f"test_file{suffix}")
    interface_class.default_interface(fp).write_file(df, fp)

    df_new, _ = interface_class.read_file(fp)
    assert isinstance(df_new, pl.DataFrame)
    assert df_new.shape == df.shape

    # same, but using write_* and read_*
    getattr(ms, f"write_{interface_class.format}")(df, fp, overwrite=True)
    df_new, _ = getattr(ms, f"read_{interface_class.format}")(fp)
    assert isinstance(df_new, pl.DataFrame)
    assert df_new.shape == df.shape

def test_stata(tmpdir):
    df = demo_dataframe("test")
    file_out = tmpdir / "test.dta"
    ms.write_dta(df, file_out)
    # StataFileInterface.default_interface(file_out).write_file(df, file_out)
    df_new, _ = read_dta(file_out)
    for col in df.columns:
        if col.startswith(("Int", "UInt")) and not col.endswith(("64")):
            assert df_new[col].dtype == pl.Int32  # Unfixable, since stata doesn't have Int8, etc
        elif col == "UInt64":
            assert df_new[col].dtype == pl.Int64  # Bugged
        elif col == "Datetime":
            assert str(df_new[col].dtype.base_type()) == "Datetime"  # Bugged
        elif col == "Date":
            assert str(df_new[col].dtype.base_type()) == "Date"
        elif col == "Categorical":
            assert df_new[col].dtype == pl.String  # Bugged
        elif col == "Boolean":
            assert df_new[col].dtype == pl.Int64  # Bugged
        elif col == "NA":
            assert df_new[col].dtype in [pl.String, pl.Int64]  # Bugged
        else:
            assert df[col].dtype == df_new[col].dtype



"""Module for the free text distribution that detects the language."""
from __future__ import annotations

from typing import Iterable, Optional

# from lingua._constant import LETTERS, PUNCTUATION
import regex
from faker import Faker
from lingua import LanguageDetectorBuilder
from scipy.stats import poisson

from metasyn.distribution.base import (
    BaseDistribution,
    BaseFitter,
    builtin_fitter,
    convert_to_series,
    metadist,
)

LETTERS = regex.compile(r"\p{Han}|\p{Hangul}|\p{Hiragana}|\p{Katakana}|\p{L}+")
PUNCTUATION = regex.compile(r"\p{P}")


@metadist(name="core.freetext", var_type="string")
class FreeTextDistribution(BaseDistribution):
    """Free text distribution.

    This distribution detects the language and generates sentences using
    the Faker package. The average number of sentences and words per item
    are detected using regexes (with the lingua package).

    Parameters
    ----------
    locale: str
        Locale used for the faker package.
    avg_sentences:
        Average number of sentences (punctuation marks) per (non-NA) row,
        if None do not make sentences.
    avg_words:
        Average number of words per (non-NA) row.

    """

    def __init__(self, locale: str, avg_sentences: Optional[float], avg_words: float):
        self.locale: str = locale
        self.avg_sentences = avg_sentences
        self.avg_words = avg_words
        self.fake = Faker(locale=self.locale)


    def draw(self):
        if self.avg_sentences is None:
            n_words = max(1, poisson(self.avg_words).rvs())
            sentence = self.fake.sentence(n_words)
            return sentence[:-1]

        n_sentences = max(1, poisson(self.avg_sentences).rvs())
        avg_words_per_sent = max(1, self.avg_words/max(1, self.avg_sentences))
        n_words = max(1, poisson(avg_words_per_sent).rvs())
        return " ".join(self.fake.sentence(nb_words=n_words) for _ in range(n_sentences))

    def information_criterion(self, values) -> float:
        series = convert_to_series(values)
        # Check the average number of characters
        avg_chars = series.str.len_chars().mean()
        if avg_chars is not None and avg_chars >= 25:  # type: ignore  # Workaround polars typing
            lang = detect_language(series)
            if lang is not None:
                return -1.0
        return 99999999

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls("en_US", 3, 6)

    def _param_dict(self):
        return {
            "locale": self.locale,
            "avg_sentences": self.avg_sentences,
            "avg_words": self.avg_words,
        }

    @classmethod
    def _param_schema(cls):
        return {
            "locale": {"type": "string"},
            "avg_sentences": {"type": ["number", "null"]},
            "avg_words": {"type": "number"},
        }

@builtin_fitter(distribution=FreeTextDistribution, var_type="string")
class FreeTextFitter(BaseFitter):
    """Fitter for the freetext distribution."""

    distribution: type[FreeTextDistribution]

    def _fit(self, series, max_values: int = 50):
        """Select the appropriate faker function and locale."""
        lang_str = detect_language(series[:max_values])
        if lang_str is None:
            return self.distribution.default_distribution()

        try:
            Faker(lang_str)
        except AttributeError:
            lang_str = "EN"

        all_text = "\n".join(series)
        n_non_empty = (series != "").sum()
        n_punctuation = len(list(PUNCTUATION.finditer(all_text)))
        n_words = len(list(LETTERS.finditer(all_text)))
        if n_punctuation < n_non_empty//3:
            avg_sentence = None
        else:
            avg_sentence = n_punctuation/len(series)
        avg_words = n_words/len(series)
        return self.distribution(lang_str, avg_sentence, avg_words)


def detect_language(values: Iterable) -> Optional[str]:
    """Detect the language of some text.

    Parameters
    ----------
    values:
        Values to detect the language of (usually polars dataframe).

    Returns
    -------
    language:
        Two letter ISO code to represent the language, or None if it could not be determined.

    """
    detector = LanguageDetectorBuilder.from_all_languages().with_low_accuracy_mode().build()
    lang = detector.detect_language_of("\n".join(values))
    if lang is None:
        return None
    return str(lang.iso_code_639_1).rsplit(".", maxsplit=1)[-1]



"""Module for testing the functionality of distributions and registries.

The testutils module provides a set of utilities for testing the functionality
and internal consistency of individual distributions and registries.
"""


from __future__ import annotations

import json
from typing import Sequence

import jsonschema
import numpy as np
import polars as pl
from jsonschema.exceptions import SchemaError

from metasyn.distribution.base import BaseDistribution, BaseFitter
from metasyn.distribution.categorical import MultinoulliDistribution
from metasyn.distribution.na import NADistribution
from metasyn.metaframe import _jsonify
from metasyn.privacy import BasePrivacy
from metasyn.registry import (
    DistributionRegistry,
)
from metasyn.var import MetaVar


def check_distribution_registry(registry_name: str):
    """Check internal consistency of a distribution registry.

    Arguments
    ---------
    registry_name:
        Name of the registry to be tested.
    """
    registry = DistributionRegistry.parse(registry_name)
    assert isinstance(registry, DistributionRegistry)
    assert len(registry.fitters) > 0
    assert all(issubclass(fitter, BaseFitter) for fitter in registry.fitters)
    assert all(issubclass(fitter.distribution, BaseDistribution) for fitter in registry.fitters)

    for fit in registry.fitters:
        n_fit = 0
        for other_fit in registry.fitters:
            if other_fit == fit:
                n_fit += 1
        assert n_fit == 1, f"Fitter {fit} exists multiple times in registry."


def check_fitter(fitter: type[BaseFitter], privacy: BasePrivacy,
                       provenance: str, test_empty: bool=True):
    """Check whether the distributions in the package can be validated positively.

    Arguments
    ---------
    distribution:
        Distribution to validate to check whether it behaves as expected.
    privacy:
        Level/type of privacy the distribution adheres to.
    provenance:
        Which registry/plugin/package provides the distribution.
    test_empty:
        If this is set to true, this will also check empty series and if the distribution
        can fit them. Otherwise, ignore testing the distribution on empty series.
    """
    # Check the schema of the distribution.
    assert issubclass(fitter, BaseFitter)
    distribution = fitter.distribution
    schema = distribution.schema()
    dist_dict = distribution.default_distribution().to_dict()
    try:
        jsonschema.validate(_jsonify(dist_dict), schema)
    except SchemaError as err:
        raise ValueError(f"Failed distribution validation for {distribution.__name__}") from err

    # Check the privacy
    assert privacy.is_compatible(fitter)
    if isinstance(distribution.var_type, str):
        var_types: Sequence[str] = [distribution.var_type]
    else:
        var_types = distribution.var_type
    for vt in var_types:
        DistributionRegistry.parse(provenance).find_fitter(
            distribution.name, var_type=vt, privacy=privacy,
            unique=distribution.unique)

    assert len(distribution.name.split(".")) == 2
    # assert distribution.provenance == provenance
    assert distribution.var_type != "unknown"
    var_types = distribution.var_type
    if isinstance(var_types, str):
        var_types = [var_types]
    for var_type in var_types:
        dist = distribution.default_distribution(var_type)
        series = pl.Series([dist.draw() for _ in range(100)])
        new_dist = fitter(privacy).fit(series)
        assert isinstance(new_dist, distribution)
        assert set(list(new_dist.to_dict())) >= set(
            ("name", "class_name", "parameters"))
        if test_empty:
            empty_series = pl.Series([], dtype=series.dtype)
            new_dist = fitter(privacy).fit(empty_series)
            assert isinstance(new_dist, distribution)




HEADER_TEMPLATE = """
# Metasyn report for {file_name}

## General

- Number of rows: {n_rows}
- Number of columns: {n_columns}
- Generated by {program_name} version {program_version} at {generation_time}

## Variables / columns

"""


VAR_TEMPLATE = """
### {var_name}

- Distribution {class_name} with parameters:
{parameters}
- Variable type {var_type}
- Based on {n_based_on} values{disclosure}
- Percentage missing: {missing_perc} %
- Generated examples: {example_list}


"""

def create_md_report(file_name, out_md_file):
    """Create markdown report from GMF file."""
    with open(file_name, "r") as handle:
        gmf_dict = json.load(handle)

    header = HEADER_TEMPLATE.format(
        file_name=file_name,
        n_rows=gmf_dict["n_rows"],
        n_columns=gmf_dict["n_columns"],
        program_name=gmf_dict["provenance"]["created by"]["name"],
        program_version=gmf_dict["provenance"]["created by"]["version"],
        generation_time=gmf_dict["provenance"]["creation time"],
    )
    variables = ""

    for var_dict in gmf_dict["vars"]:
        var = MetaVar.from_dict(var_dict)
        if isinstance(var.distribution, MultinoulliDistribution):
            parameters = [f"\t- {label}: {round(prob*gmf_dict['n_rows']*(1-var.prop_missing))}\n"
                          for label, prob in zip(var.distribution.labels, var.distribution.probs)]
        elif isinstance(var.distribution, NADistribution):
            variables += (f"### {var.name}\n- Distribution NADistribution\n- Only missing values"
                          "\n- Examples: NA, NA, NA, ...\n")
            continue
        else:
            parameters = [f"\t - {name}: {value}\n"
                             for name, value in var.distribution._param_dict().items()]
        parameter_str = "".join(parameters)
        if var.prop_missing > 0:
            examples = np.random.permutation([str(var.distribution.draw()) for _ in range(3)] +
                                             ["NA", "NA"])
        else:
            examples = [str(x) for x in var.draw_series(5, None)]

        if "privacy" in var_dict["creation_method"]:
            partition_size = var_dict["creation_method"]["privacy"]["parameters"]["partition_size"]
            disclosure = f" using micro aggregation with a partition size of {partition_size}"
        else:
            disclosure = ""
        variables += VAR_TEMPLATE.format(
            var_name = var.name,
            var_type=var.var_type,
            class_name=var.distribution.__class__.__name__,
            n_based_on=round(gmf_dict["n_rows"]*(1-var.prop_missing)),
            example_list=", ".join(examples) + ", ...",
            parameters=parameter_str,
            missing_perc=f"{100*var.prop_missing:.2f}",
            disclosure=disclosure,
        )
    with open(out_md_file, "w", encoding="utf-8") as handle:
        handle.write(header + variables)


def create_input_toml(file_name):
    """Create input toml with all distribution in builtin."""
    import tomlkit  # noqa: PLC0415

    doc = tomlkit.document()
    doc.add("config_version", "1.1")
    doc.add("plugins", ["builtin"])
    doc.add("n_rows", 100)
    doc.add("defaults", {"data_free": True, "prop_missing": 0.1})
    var_array = tomlkit.aot()
    for dist in DistributionRegistry.parse("builtin").distributions:
        var = tomlkit.table()
        var.add("name", dist.__name__)
        if isinstance(dist.var_type, str):
            var_type = dist.var_type
        else:
            var_type = dist.var_type[0]
        var.add("var_type", var_type)
        dist_dict = _jsonify(dist.default_distribution().to_dict())
        dist_dict.pop("version")
        dist_dict.pop("class_name")
        var.add("distribution", dist_dict)
        var.add(tomlkit.nl())
        var_array.append(var)
    doc.add("var", var_array)

    with open(file_name, "w", encoding="utf-8") as handle:
        tomlkit.dump(doc, handle)



"""Module with the exponential distribution."""

from scipy.stats import expon

from metasyn.distribution.base import (
    BaseDistribution,
    BaseFitter,
    ScipyDistribution,
    builtin_fitter,
    metadist,
)


@metadist(name="core.exponential", var_type="continuous")
class ExponentialDistribution(ScipyDistribution):
    """Exponential distribution for floating point type.

    This class implements the exponential distribution with the rate as its
    single parameter.

    Parameters
    ----------
    rate:
        Rate of the exponential distribution. This is equal to 1/mean of the distribution.

    Examples
    --------
    >>> ExponentialDistribution(rate=2.4)
    """

    scipy_class = expon

    def __init__(self, rate: float):
        self.par = {"rate": rate}
        self.dist = expon(loc=0, scale=1/max(rate, 1e-8))

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls(1.0)

    @classmethod
    def _param_schema(cls):
        return {
            "rate": {"type": "number"}
        }


@builtin_fitter(distribution=ExponentialDistribution, var_type="continuous")
class ExponentialFitter(BaseFitter):
    """Fitter for exponential distribution."""

    def _fit(self, series):
        series = series.filter(series > 0)
        if len(series) == 0:
            return self.distribution.default_distribution()
        return self.distribution(rate=1/expon.fit(series, floc=0)[1])



"""Uniform distributions and fitters."""

import datetime as dt
from abc import abstractmethod
from random import random
from typing import Any, Dict

import numpy as np
from scipy.stats import randint, uniform

from metasyn.distribution.base import (
    BaseDistribution,
    BaseFitter,
    ScipyDistribution,
    builtin_fitter,
    metadist,
)
from metasyn.distribution.util import convert_numpy_datetime


@metadist(name="core.uniform", var_type="discrete")
class DiscreteUniformDistribution(ScipyDistribution):
    """Uniform discrete distribution.

    It differs from the floating point uniform distribution by
    being a discrete distribution instead.

    Parameters
    ----------
    lower: int
        Lower bound (inclusive) of the uniform distribution.
    upper: int
        Upper bound (exclusive) of the uniform distribution.

    Examples
    --------
    >>> DiscreteUniformDistribution(lower=3, upper=20)
    """

    # dist = randint

    def __init__(self, lower: int, upper: int):
        self.par = {"lower": lower, "upper": upper}
        self.dist = randint(low=lower, high=upper)

    def _information_criterion(self, values):
        return np.log(len(values))*self.n_par - 2*np.sum(self.dist.logpmf(values))


    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls(0, 10)

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "integer"},
            "upper": {"type": "integer"},
        }


@builtin_fitter(distribution=DiscreteUniformDistribution, var_type="discrete")
class DiscreteUniformFitter(BaseFitter):
    """Fitter for discrete uniform distribution."""

    def _fit(self, series):
        return DiscreteUniformDistribution(series.min(), series.max()+1)




@metadist(name="core.uniform", var_type="continuous")
class ContinuousUniformDistribution(ScipyDistribution):
    """Uniform distribution for floating point type.

    This class implements the uniform distribution between a minimum
    and maximum.

    Parameters
    ----------
    lower: float
        Lower bound for uniform distribution.
    upper: float
        Upper bound for uniform distribution.

    Examples
    --------
    >>> UniformDistribution(lower=-3.0, upper=10.0)
    """

    # dist = uniform

    def __init__(self, lower: float, upper: float):
        self.par = {"lower": lower, "upper": upper}
        self.dist = uniform(loc=self.lower, scale=max(self.upper-self.lower, 1e-8))

    def _information_criterion(self, values):
        if np.any(np.array(values) < self.lower) or np.any(np.array(values) > self.upper):
            return np.log(len(values))*self.n_par + 100*len(values)
        if np.fabs(self.upper-self.lower) < 1e-8:
            return np.log(len(values))*self.n_par - 100*len(values)
        return (np.log(len(values))*self.n_par
                - 2*len(values)*np.log((self.upper-self.lower)**-1))

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls(0, 1)

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "number"},
            "upper": {"type": "number"},
        }

@builtin_fitter(distribution=ContinuousUniformDistribution, var_type="continuous")
class ContinuousUniformFitter(BaseFitter):
    """Fitter for continuous uniform distribution."""

    def _fit(self, series):
        return ContinuousUniformDistribution(series.min(), series.max())


class BaseDTUniformDistribution(BaseDistribution):
    """Base class for all time/date/datetime uniform distributions."""

    precision_possibilities = ["microseconds", "seconds", "minutes", "hours", "days"]

    def __init__(self, lower: Any, upper: Any, precision: str = "microseconds"):
        if isinstance(lower, str):
            lower = self.fromisoformat(lower)
        elif isinstance(lower, np.datetime64):
            lower = convert_numpy_datetime(lower)
        if isinstance(upper, str):
            upper = self.fromisoformat(upper)
        elif isinstance(upper, np.datetime64):
            upper = convert_numpy_datetime(upper)
        self.precision = precision
        self.lower = self.round(lower)
        self.upper = self.round(upper)


    def round(self, time_obj: Any) -> Any:
        """Round down any time object with the precision.

        Parameters
        ----------
        time_obj:
            Object to round down.

        Return
        ------
        obj:
            Time/date/datetime object rounded down to the measured precision.
        """
        for prec in self.precision_possibilities:
            if prec == self.precision:
                break
            time_obj = time_obj.replace(**{prec[:-1]: 0})
        try:
            time_obj = time_obj.replace(nanosecond=0)
        except TypeError:
            pass
        return time_obj

    def draw(self) -> dt.datetime:
        delta = self.upper-self.lower + self.minimum_delta
        new_time = random()*delta + self.lower
        return self.round(new_time)

    @abstractmethod
    def fromisoformat(self, dt_obj: str):
        """Convert string to iso format."""

    @property
    def minimum_delta(self) -> dt.timedelta:
        """Get the minimum time delta."""
        return dt.timedelta(**{self.precision: 1})

    def information_criterion(self, values): # noqa: ARG002
        return 0.0

    def _param_dict(self):
        return {
            "lower": self.lower.isoformat(),
            "upper": self.upper.isoformat(),
            "precision": self.precision,
        }



@metadist(name="core.uniform", var_type="datetime")
class DateTimeUniformDistribution(BaseDTUniformDistribution):
    """Uniform DateTime distribution.

    Datetime objects will be uniformly distributed between a start and end date time.

    Parameters
    ----------
    lower: str or datetime.datetime
        Earliest possible datetime to be generated from this distribution.
        String formatted in ISO format, see examples below or use datetime objects
        from the python standard library `datetime`.
    upper: str or datetime.datetime
        Latest possible datetime to be generated from this distribution.
        String formatted in ISO format, see examples below or use datetime objects
        from the python standard library `datetime`.
    precision: str
        Most precise measure of datetime that the distribution will take into account.
        For example, if precision == "seconds", then the microseconds will always be
        set to 0. Possible values: ["microseconds", "seconds", "minutes", "hours", "days"].

    Examples
    --------
    >>> DateTimeUniformDistribution(lower="2022-07-15T10:39", upper="2022-08-15T10:39",
                                    precision="minutes")
    """

    def fromisoformat(self, dt_obj: str) -> dt.datetime:
        return dt.datetime.fromisoformat(dt_obj)

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls("2022-07-15T10:39:36", "2022-08-15T10:39:36", precision="seconds")

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "string"},
            "upper": {"type": "string"},
            "precision": {"type": "string"},
        }


@metadist(name="core.uniform", var_type="time")
class TimeUniformDistribution(BaseDTUniformDistribution):
    """Uniform time distribution.

    Time objects will be uniformly distributed between a start and end time.

    Parameters
    ----------
    lower: str or datetime.time
        Earliest possible time to be generated from this distribution.
        String formatted in ISO format, see examples below or use time objects
        from the python standard library `datetime`.
    upper: str or datetime.time
        Latest possible time to be generated from this distribution.
        String formatted in ISO format, see examples below or use time objects
        from the python standard library `datetime`.
    precision: str
        Most precise measure of time that the distribution will take into account.
        For example, if precision == "seconds", then the microseconds will always be
        set to 0. Possible values: ["microseconds", "seconds", "minutes", "hours"].

    Examples
    --------
    >>> TimeUniformDistribution(lower="10:39:12", upper="10:39:45", precision="seconds")
    """

    def fromisoformat(self, dt_obj: str) -> dt.time:
        return dt.time.fromisoformat(dt_obj)

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls("10:39:36", "18:39:36", precision="seconds")

    def draw(self):
        dt_lower = dt.datetime.combine(dt.datetime.today(), self.lower)
        dt_upper = dt.datetime.combine(dt.datetime.today(), self.upper)
        delta = dt_upper-dt_lower + self.minimum_delta
        return self.round((random()*delta + dt_lower).time())

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "string"},
            "upper": {"type": "string"},
            "precision": {"type": "string"},
        }


@metadist(name="core.uniform", var_type="date")
class DateUniformDistribution(BaseDTUniformDistribution):
    """Uniform date distribution.

    Date objects will be uniformly distributed between a start and end date.

    Parameters
    ----------
    lower: str or datetime.date
        Earliest possible date to be generated from this distribution.
        String formatted in ISO format, see examples below or use date objects
        from the python standard library `datetime`.
    upper: str or datetime.date
        Latest possible date to be generated from this distribution.
        String formatted in ISO format, see examples below or use date objects
        from the python standard library `datetime`.

    Examples
    --------
    >>> DateUniformDistribution(lower="10:39:12", upper="10:39:45", precision="seconds")
    """

    precision_possibilities = ["days"]

    def __init__(self, lower: Any, upper: Any):
        super().__init__(lower, upper, precision="days")

    def fromisoformat(self, dt_obj: str) -> dt.date:
        return dt.date.fromisoformat(dt_obj)

    def round(self, time_obj):
        return time_obj

    def _param_dict(self) -> Dict:
        date_dict = BaseDTUniformDistribution._param_dict(self)
        del date_dict["precision"]
        return date_dict

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls("1903-07-15", "1940-07-16")

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "string"},
            "upper": {"type": "string"},
        }


class BaseDTUniformFitter(BaseFitter):
    """Base class for date/time/datetime uniform distributions."""

    precision_possibilities = ["microseconds", "seconds", "minutes", "hours", "days"]

    @classmethod
    def _get_precision(cls, values):
        cur_precision = 0
        for precision in cls.precision_possibilities[:-1]:
            if not np.all([getattr(d, precision[:-1]) == 0 for d in values]):
                break
            cur_precision += 1
        return cls.precision_possibilities[cur_precision]


@builtin_fitter(distribution=TimeUniformDistribution, var_type="time")
class TimeUniformFitter(BaseDTUniformFitter):
    """Fitter for time uniform distribution."""

    # precision_possibilities = ["microseconds", "seconds", "minutes", "hours", "days"]
    def _fit(self, values):
        return TimeUniformDistribution(values.min(), values.max(), self._get_precision(values))

@builtin_fitter(distribution=DateTimeUniformDistribution, var_type="datetime")
class DateTimeUniformFitter(BaseDTUniformFitter):
    """Fitter for datetime uniform distribution."""

    def _fit(self, values):
        return DateTimeUniformDistribution(values.min(), values.max(), self._get_precision(values))


@builtin_fitter(distribution=DateUniformDistribution, var_type="date")
class DateUniformFitter(BaseDTUniformFitter):
    """Fitter for date uniform distribution."""

    precision_possibilities = ["days"]
    def _fit(self, values):
        return DateUniformDistribution(values.min(), values.max())



"""Unique key distributions, often for primary key columns."""

from typing import Set, TypeVar

import numpy as np

from metasyn.distribution.base import (
    BaseDistribution,
    BaseFitter,
    builtin_fitter,
    convert_to_series,
    metadist,
)

UKeyT = TypeVar("UKeyT", bound="UniqueKeyDistribution")


@metadist(name="core.unique_key", var_type="discrete", unique=True)
class UniqueKeyDistribution(BaseDistribution):
    """Unique key distribution for identifiers.

    Discrete distribution that ensures the uniqueness of the drawn values.

    Parameters
    ----------
    lower:
        Minimum value for the keys.
    consecutive:
        True if keys are consecutive and increasing, False otherwise.

    Examples
    --------
    >>> UniqueKeyDistribution(lower=0, consecutive=True)
    """

    def __init__(self, lower: int, consecutive: bool):
        self.par = {"lower": lower, "consecutive": consecutive}
        self.last_key = lower - 1
        self.key_set: Set[int] = set()

    def _param_dict(self):
        return self.par

    def __getattr__(self, attr: str):
        """Get attribute for easy access to parameters.

        Parameters
        ----------
        attr:
            Attribute to retrieve. If the attribute is a parameter
            name, then retrieve that parameter, otherwise use the default
            implementation for getting the attribute.

        Returns
        -------
        object:
            Parameter or attribute.
        """
        if attr != "par" and attr in self.par:
            return self.par[attr]
        return object.__getattribute__(self, attr)

    def draw_reset(self):
        self.last_key = self.lower - 1
        self.key_set = set()

    def draw(self):
        if self.consecutive == 1:
            self.last_key += 1
            return self.last_key

        while True:
            random_number = np.random.randint(self.lower,
                                              self.lower+2*len(self.key_set)+2)
            if random_number not in self.key_set:
                self.key_set.add(random_number)
                return random_number

    def information_criterion(self, values):
        values = convert_to_series(values)
        if len(values) == 0:
            return 2 * 2

        if values.min() < self.lower:
            return 2*np.log(len(values))+999*len(values)

        # If the values are not unique the fit is extremely bad.
        if len(set(values)) != len(values):
            return 2*np.log(len(values))+999*len(values)

        lower = values.min()
        high = values.max()+1

        if self.consecutive == 1:
            # Check if the values are truly consecutive
            if len(values) == high-lower and np.all(values.to_numpy() == np.arange(lower, high)):
                return 2*np.log(len(values))
            return 2*np.log(len(values))+999*len(values)

        n_choice = high - lower

        # Probabilities go up like 1/n, 1/(n-1), 1/(n-2), ..., 1/2, 1
        return (3*np.log(len(values))
                - 2*np.sum(np.log(1/np.arange(n_choice, n_choice-len(values), -1))))

    @classmethod
    def default_distribution(cls: type[UKeyT], var_type=None) -> UKeyT: # noqa: ARG003
        return cls(0, False)

    @classmethod
    def _param_schema(cls):
        return {
            "lower": {"type": "integer"},
            "consecutive": {"type": "boolean"},
        }

@builtin_fitter(distribution=UniqueKeyDistribution, var_type="discrete")
class UniqueKeyFitter(BaseFitter):
    """Fitter for unique key distribution."""

    distribution: type[UniqueKeyDistribution]

    def _fit(self, series) -> UniqueKeyDistribution:
        lower = series.min()
        high = series.max() + 1
        if len(series) == high-lower and np.all(series.to_numpy() == np.arange(lower, high)):
            return self.distribution(lower, True)
        return self.distribution(lower, False)



"""Utilities for builtin distributions and fitters."""

import datetime as dt

import numpy as np


def convert_numpy_datetime(time_obj: np.datetime64) -> dt.datetime:
    """Convert numpy datetime to python stdlib datetime.

    Parameters
    ----------
    time_obj:
        datetime to be converted.

    Returns
    -------
    datetime.datetime:
        Converted datetime.
    """
    unix_epoch = np.datetime64(0, 's')
    one_second = np.timedelta64(1, 's')
    seconds_since_epoch = (time_obj - unix_epoch) / one_second
    return dt.datetime.fromtimestamp(float(seconds_since_epoch))




"""Module for testing parsing configurations (from file or otherwise)."""

from pathlib import Path

import pytest
from pytest import mark

from metasyn.config import MetaConfig, VarSpecAccess
from metasyn.distribution.uniform import ContinuousUniformDistribution
from metasyn.privacy import BasePrivacy
from metasyn.varspec import DistributionSpec, VarSpec


@mark.parametrize(
    "input,error",
    [
        ("uniform", False),
        (None, False),
        ({"name": {"uniform": False}}, False),
        (ContinuousUniformDistribution, False),
        (ContinuousUniformDistribution(0, 2), False),
        (DistributionSpec(), False),
        ({"fit_kwargs": {"param": 3}}, True),
        ({"version": "2.0"}, True),
        ({"parameters": {"param": 2}}, True),
        (1, True),
    ]
)
def test_dist_spec(input, error):
    """Test parsing distribution parameters for different inputs."""
    if error:
        with pytest.raises(Exception):
            DistributionSpec.parse(input)
    else:
        dist_spec = DistributionSpec.parse(input)
        assert isinstance(dist_spec, DistributionSpec)

def test_var_spec():
    """Test various ways to initialize the VarSpec class."""
    var_spec = VarSpec("test", privacy={"name": "none", "parameters": {}})
    assert var_spec.name == "test"
    assert isinstance(var_spec.privacy, BasePrivacy)
    with pytest.raises(ValueError):
        var_spec = VarSpec("test", data_free=True)
    var_spec = VarSpec("test")
    assert isinstance(var_spec.dist_spec, DistributionSpec)
    var_spec = VarSpec.from_dict({"name": "test"})
    assert var_spec.name == "test"
    var_spec = VarSpec.from_dict({"name": "test", "distribution": "uniform"})
    assert var_spec.dist_spec.name == "uniform"


def test_meta_config_datafree():
    """Test for creating a MetaConfig class from a .toml file without data."""
    meta_config = MetaConfig.from_toml(Path("tests", "data", "no_data_config.toml"))
    assert meta_config.n_rows == 100
    assert len(meta_config.var_specs) == 3
    assert isinstance(meta_config.var_specs[0], VarSpec)
    assert meta_config.var_specs[0].privacy is None
    assert isinstance(meta_config.var_specs[0].dist_spec, DistributionSpec)
    assert isinstance(meta_config.to_dict(), dict)
    var_spec = meta_config.get("PassengerId")
    assert isinstance(var_spec, VarSpecAccess)

    assert var_spec.data_free is True
    var_spec = meta_config.get("unknown")
    assert var_spec.name == "unknown"

    all_var_spec = list(meta_config.iter_var())
    assert len(all_var_spec) == 3
    assert isinstance(all_var_spec[0], VarSpecAccess)
    assert all_var_spec[0].meta_config == meta_config
    assert len(list(meta_config.iter_var(exclude=["PassengerId"]))) == 2


def test_meta_config():
    """Test the creation of a MetaConfig class that is not data free."""
    with pytest.raises(ValueError):
        MetaConfig.from_toml(Path("tests", "data", "titanic.csv"))
    meta_config = MetaConfig.from_toml(Path("tests", "data", "example_config.toml"))
    assert len(meta_config.var_specs) == 5
    var_spec = meta_config.get("Cabin")
    assert var_spec.data_free is False
    assert var_spec.var_type is None
    assert var_spec.dist_spec.name == "core.regex"



"""Module for distributions that use the faker package."""
from __future__ import annotations

from typing import Iterable

# from lingua._constant import LETTERS, PUNCTUATION
from faker import Faker

from metasyn.distribution.base import (
    BaseDistribution,
    BaseFitter,
    UniqueDistributionMixin,
    builtin_fitter,
    metadist,
)


@metadist(name="core.faker", var_type="string")
class FakerDistribution(BaseDistribution):
    """Faker distribution for cities, addresses, etc.

    This is mainly an interface for the faker package, so that it
    can be used within the metasyn package. It doesn't have any
    true fitting/statistical inference method, so it has to be manually
    selected.

    Parameters
    ----------
    faker_type: str
        The provider function in the faker package, e.g. 'city' or 'ipv4', etc.
    locale: str
        Locale used for the faker package.

    Examples
    --------
    >>> FakerDistribution(faker_type="city", locale="en_US")
    >>> FakerDistribution(faker_type="address", locale="nl_NL")

    """

    def __init__(self, faker_type: str, locale: str = "en_US"):
        self.faker_type: str = faker_type
        self.locale: str = locale
        self.fake: Faker = Faker(locale=locale)

    def draw(self):
        return getattr(self.fake, self.faker_type)()

    def information_criterion(self, values: Iterable) -> float: # noqa: ARG002
        return 99999

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls("city")

    def _param_dict(self):
        return {
            "faker_type": self.faker_type,
            "locale": self.locale
        }

    @classmethod
    def _param_schema(cls):
        return {
            "faker_type": {"type": "string"},
            "locale": {"type": "string"},
        }


@builtin_fitter(distribution=FakerDistribution, var_type="string")
class FakerFitter(BaseFitter):
    """Fitter for the faker distribution."""

    distribution: type[FakerDistribution]

    def _fit(self, values, faker_type: str = "city", locale: str = "en_US"): # noqa: ARG002
        """Select the appropriate faker function and locale."""
        return self.distribution(faker_type, locale)


@metadist(name="core.faker", var_type="string")
class UniqueFakerDistribution(UniqueDistributionMixin, FakerDistribution):
    """Faker distribution that returns unique values.

    See :class:`~FakerDistribution` for examples and explanation.
    """

@builtin_fitter(distribution=UniqueFakerDistribution, var_type="string")
class UniqueFakerFitter(FakerFitter):
    """Fitter for the unique faker distribution."""




"""Metasyn: a package for creating synthetic datasets.

Metasyn has three main purposes:

1. Estimation: Metasyn can create a MetaFrame from a dataset.
A MetaFrame is metadata describing a table, augmented with statistical
information on the columns. It captures individual distributions and
features and enables generation of synthetic data based on it.

2. Serialization and deserialization: Metasyn can save a
MetaFrame into an easy to read GMF file. This allows users to audit,
understand, and modify their data generation model. These GMF files
can also be imported back into Metasyn to generate synthetic data.

3. Generation: Metasyn can generate synthetic data based on a MetaFrame.
The synthetic data produced solely depends on the MetaFrame, thereby
maintaining a critical separation between the original sensitive data and the
generated synthetic data.
"""

from importlib.metadata import version

from metasyn import distribution, file, privacy
from metasyn.demo.dataset import demo_dataframe, demo_file
from metasyn.distribution.base import metadist
from metasyn.file import (
    read_csv,
    read_dta,
    read_excel,
    read_sav,
    read_tsv,
    write_csv,
    write_dta,
    write_excel,
    write_sav,
    write_tsv,
)
from metasyn.metaframe import MetaFrame
from metasyn.var import MetaVar
from metasyn.varspec import VarSpec

__all__ = [
    "MetaVar",
    "MetaFrame",
    "demo_file",
    "demo_dataframe",
    "metadist",
    "VarSpec",
    "read_csv",
    "read_sav",
    "read_tsv",
    "read_excel",
    "read_dta",
    "write_csv",
    "write_dta",
    "write_excel",
    "write_sav",
    "write_tsv",
    "distribution",
    "file",
    "privacy",
]
__version__ = version("metasyn")



"""Module implementing the distribution registry.

Distribution registries are used to find/fit distributions that are available.
See pyproject.toml on how the builtin distributions are registered.
"""

from __future__ import annotations

import warnings
from importlib.metadata import entry_points
from inspect import signature
from typing import TYPE_CHECKING, Any, Optional, Union

import numpy as np
import polars as pl

from metasyn.distribution.base import BaseDistribution, BaseFitter
from metasyn.distribution.na import NADistribution
from metasyn.privacy import BasePrivacy, BasicPrivacy
from metasyn.util import get_registry
from metasyn.varspec import DistributionSpec

if TYPE_CHECKING:
    from metasyn.config import VarSpec, VarSpecAccess


class DistributionRegistry():
    """Registry of distributions and fitters.

    This class is responsible for managing and providing access to
    fitters and distributions. It allows for fitting distributions,
    as well as retrieving distributions/fitters based on certain constraints
    such as privacy level, variable type, and uniqueness.

    You can directly initialize the class with a list of fitters, but most likely
    you will want to use the :meth:`DistributionRegistry.parse` method, which can load
    fitters from registries provided by plugins.

    Parameters
    ----------
    fitters:
        Fitters to initialize the registry with.
    """

    def __init__(
            self,
            fitters: list[type[BaseFitter]]):
        self.fitters = fitters

    @classmethod
    def parse(cls, plugins: Union[list[str], None, str]):
        """Initialize the distribution registry from plugin names.

        Parameters
        ----------
        plugins:
            Name of plugin(s) for fitters/distribution or a list of names.
        """
        fitters = []
        if isinstance(plugins, str):
            plugins = [plugins]

        entries = {e.name: e for e in entry_points(group="metasyn.distribution_registry")}
        if plugins is None:
            plugins = list(entries)

        for registry_name in plugins:
            if registry_name not in entries:
                registry = get_registry()
                if registry_name not in registry:
                    raise ValueError(
                        f"Cannot find plugin with name '{registry_name}'.")
                raise ValueError(
                    f"Plugin '{registry_name}' is not installed.\n"
                    f"See {registry[registry_name]['url']} for installation instructions."
                )
            try:
                fitters.extend(entries[registry_name].load())
            except Exception as exc:
                warnings.warn(f"Could not load plugin with name {registry_name}, plugin might be"
                              f" broken or out of date: {exc}")
        return cls(fitters)

    def fit(self, series: pl.Series,
            var_type: str,
            dist_spec: DistributionSpec,
            privacy: BasePrivacy = BasicPrivacy()) -> tuple[BaseDistribution, Optional[BaseFitter]]:
        """Fit a distribution to a column/series.

        Parameters
        ----------
        series:
            The data to fit the distributions to.
        var_type:
            The variable type of the data.
        dist_spec:
            Distribution to fit. If not supplied or None, the information
            criterion will be used to determine which distribution is the most
            suitable. For most variable types, the information criterion is based on
            the BIC (Bayesian Information Criterion).
        privacy:
            Level of privacy that will be used in the fit.
        """
        if dist_spec.distribution is not None:
            return dist_spec.distribution, None
        if dist_spec.name is not None:
            return self._fit_distribution(series, dist_spec, var_type, privacy)
        return self._find_best_fit(series, var_type, dist_spec.unique, privacy)

    def create(self, var_spec: Union[VarSpec, VarSpecAccess]) -> BaseDistribution:
        """Create a distribution without any data.

        Parameters
        ----------
        var_spec
            A variable configuration that provides all the information to create the distribution.

        Returns
        -------
            A distribution according to the variable specifications.
        """
        dist_spec = var_spec.dist_spec
        unique = dist_spec.unique if dist_spec.unique else False
        if dist_spec.name is None:
            raise ValueError("Cannot create distribution without specifying the 'name' key.")
        dist_class = self.find_distribution(
            dist_spec.name, var_spec.var_type, unique=unique)
        try:
            return dist_class(**dist_spec.parameters)  # type: ignore
        except TypeError as err:
            dist_param = set(signature(dist_class.__init__).parameters) - {"self"}  # type: ignore
            unknown_param = set(dist_spec.parameters) - dist_param  # type: ignore
            missing_param = dist_param - set(dist_spec.parameters)  # type: ignore
            if len(unknown_param) > 0:
                raise TypeError(f"Unknown parameters {unknown_param} for variable {var_spec.name}."
                                f"Available parameters: {dist_param}")
            if len(missing_param) > 0:
                raise ValueError(f"Missing parameters for variable {var_spec.name}:"
                                 f" {missing_param}.")
            raise err

    def _find_best_fit(self, series: pl.Series, var_type: str,
                       unique: Optional[bool],
                       privacy: BasePrivacy) -> tuple[BaseDistribution, Optional[BaseFitter]]:
        """Fit a distribution to a series.

        Search for the distribution within all available distributions in the tree.

        Parameters
        ----------
        series:
            Series to fit a distribution to.
        var_type:
            Variable type of the series.
        unique:
            Whether the variable should be unique or not.
        privacy:
            Privacy level to find the best fit with.

        Returns
        -------
        BaseDistribution:
            Distribution fitted to the series.
        """
        if len(series.drop_nulls()) == 0:
            return NADistribution(), None
        try_unique = unique is True
        fitters = self.filter_fitters(privacy=privacy, var_type=var_type, unique=try_unique)
        if len(fitters) == 0:
            raise ValueError(f"No available distributions with variable type: '{var_type}'"
                             f" and unique={try_unique}")
        fit_instances = [f(privacy) for f in fitters]
        dist_instances = [(f.fit(series), f) for f in fit_instances]
        dist_bic = [d.information_criterion(series) for d, _ in dist_instances]
        if unique is None:
            fit_list_unq = self.filter_fitters(privacy=privacy, var_type=var_type, unique=True)
            if len(fit_list_unq) > 0:
                fit_inst_unq = [f(privacy) for f in fit_list_unq]
                dist_inst_unq = [(f.fit(series), f) for f in fit_inst_unq]
                dist_bic_unq = [d.information_criterion(series) for d, _ in dist_inst_unq]
                # We don't want to warn about potential uniqueness too easily
                # The offset is a heuristic that ensures about 12 rows are needed for uniqueness
                # Or 5 rows for consecutive values.
                if np.min(dist_bic_unq) + 16 < np.min(dist_bic):
                    best_dist, best_fitter = dist_inst_unq[np.argmin(dist_bic_unq)]
                    if best_dist.name == "core.unique_key" and best_dist.consecutive:  # type: ignore
                        return best_dist, best_fitter
                    warnings.warn(
                        f"\nMetasyn detected that variable '{series.name}' is potentially unique.\n"
                        f"Use var_spec=[VarSpec(\"{series.name}\", unique=True)] to make it unique."
                        f"\nTo dismiss this warning use [VarSpec(\"{series.name}\", unique=False)]."
                        "\nIf you are using a configuration file add distribution = {unique = True}"
                        f" for the variable with name '{series.name}'.",
                        UserWarning
                    )

        return dist_instances[np.argmin(dist_bic)]

    def find_distribution(
            self,
            dist_name: str,
            var_type: Optional[str],
            unique: bool = False,
            version: Optional[str] = None
        ) -> type[BaseDistribution]:
        dist_classes = self.filter_distributions(name=dist_name, var_type=var_type,
                                                 unique=unique, version=version)
        if len(dist_classes) == 1:
            return dist_classes[0]

        if len(dist_classes) > 1:
            dist_str = [f"({d.__name__}, {d.var_type}, {d.unique}, {d.version})"
                        for d in dist_classes]
            raise ValueError(f"Multiple valid distributions found with name {dist_name}, var_type "
                             f"{var_type}, unique {unique}, version {version}."
                             f" Alternatives: {dist_str}")
        name_classes = self.filter_distributions(name=dist_name)
        if len(name_classes) == 0:
            raise ValueError(f"No known distributions with name '{dist_name}'.")
        dist_str = [f"({d.__name__}, {d.var_type}, {d.unique}, {d.version})"
            for d in name_classes]
        raise ValueError(f"No distribution found with name {dist_name}, var_type "
                         f"{var_type}, unique {unique}, version {version}."
                         f" Alternatives: {dist_str}")

    def find_fitter(self,
                    dist_name: str,
                    var_type: Optional[str],
                    privacy: Optional[BasePrivacy] = BasicPrivacy(),
                    unique: bool = False,
                    version: Optional[str] = None) -> type[BaseFitter]:
        """Find a distribution and fit keyword arguments from a name.

        Sometimes there might be multiple possible fitters that satisfy the criteria.
        In this case the first in the registry will be chosen. If you do not want this
        behavior, it is recommended to specify the fitter name directly.

        Parameters
        ----------
        dist_name:
            Name of the distribution that needs to be fit, e.g., for the built-in
            uniform distribution: "uniform", "core.uniform"
            or name of the fitter: "ContinuousUniformFitter".
        privacy:
            Type of privacy to be applied.
        var_type:
            Type of the variable to find. If var_type is None, then do not check the
            variable type.
        unique:
            Whether the distribution to be found is unique.
        version:
            Version of the distribution to get. If necessary get them from legacy.

        Returns
        -------
        tuple[Type[BaseFitter]:
            Fitter that satisfies the requirements.
        """
        fitter_classes = self.filter_fitters(
            name=dist_name, privacy=privacy, var_type=var_type, unique=unique, version=version)
        if len(fitter_classes) == 1:
            return fitter_classes[0]

        if len(fitter_classes) > 1:
            if var_type is None and not all([
                    f.var_type == fitter_classes[0] for f in fitter_classes]):
                raise ValueError(f"Multiple valid fitters found with name {dist_name}, "
                                 "please specify var_type.")
            return fitter_classes[0]

        name_classes = self.filter_fitters(name=dist_name)
        if len(name_classes) == 0:
            raise ValueError(f"No known fitters with name '{dist_name}'.")
        fitter_str = [f"({f.__name__}, {f.var_type}, {f.distribution.unique}, {f.version},"
                      f" {f.privacy_type})" for f in name_classes]
        raise ValueError(f"No fitter found with name {dist_name}, var_type "
                         f"{var_type}, unique {unique}, version {version}."
                         f" Alternatives: {fitter_str}")

    def _fit_distribution(self, series: pl.Series,
                          dist_spec: DistributionSpec,
                          var_type: str,
                          privacy: BasePrivacy) -> tuple[BaseDistribution, Optional[BaseFitter]]:
        """Fit a specific distribution to a series.

        In contrast the fit method, this needs a supplied distribution(type).

        Parameters
        ----------
        series:
            Series to fit the distribution to.
        dist_spec:
            Distribution to fit (if it is not already fitted).
        var_type:
            Type of variable to fit the distribution for.
        privacy:
            Privacy level to fit the distribution with.

        Returns
        -------
        BaseDistribution:
            Fitted distribution.
        """
        unique = dist_spec.unique
        unique = unique if unique else False
        assert dist_spec.name is not None

        # If the parameters are already specified, the privacy level doesn't matter anymore.
        if dist_spec.parameters is not None:
            dist_class = self.find_distribution(dist_spec.name, var_type, unique=unique)
            return dist_class(**dist_spec.parameters), None

        fitter_class = self.find_fitter(dist_spec.name, var_type, privacy=privacy,
                                        unique=unique)

        fit_kwargs = dist_spec.fit_kwargs
        fitter = fitter_class(privacy)
        dist_instance = fitter.fit(series, **fit_kwargs)
        return dist_instance, fitter

    def filter_fitters(self,
                       name: Optional[str] = None,
                       privacy: Optional[BasePrivacy] = None,
                       var_type: Optional[str] = None,
                       unique: bool = False,
                       version: Optional[str] = None) -> list[type[BaseFitter]]:
        """Get the available distributions with constraints.

        Parameters
        ----------
        privacy:
            Privacy level/type to filter the distributions.
        var_type:
            Variable type to filter for, e.g. 'string'.
        unique:
            Whether the distributions to be gotten are unique.
        use_legacy:
            Whether to use legacy distributions or not.

        Returns
        -------
        dist_list:
            List of distributions that fit the given constraints.
        """
        fitters = self.fitters
        if name is not None:
            fitters = [f for f in fitters if f.matches_name(name)]
        if var_type is not None:
            fitters = [f for f in fitters if f.provides_var_type(var_type)]
        fitters = [f for f in fitters if f.distribution.unique == unique]
        if privacy is not None:
            fitters = [f for f in fitters if f.privacy_type == privacy.name]
        if version is not None:
            fitters = [f for f in fitters if f.version == version]
        return fitters

    def filter_distributions(self, name: Optional[str] = None, var_type: Optional[str] = None,
                             unique: Optional[bool] = False, version: Optional[str] = None):
        dist = self.distributions

        if name is not None:
            dist = [d for d in dist if d.matches_name(name)]

        if var_type is not None:
            dist = [d for d in dist if d.provides_var_type(var_type)]

        if unique is not None:
            dist = [d for d in dist if d.unique == unique]

        if version is not None:
            dist = [d for d in dist if d.version == version]

        return dist

    def from_dict(self, var_dict: dict[str, Any]) -> BaseDistribution:
        """Create a distribution from a dictionary.

        Parameters
        ----------
        var_dict:
            Variable dictionary that includes the distribution properties.

        Returns
        -------
        BaseDistribution:
            Distribution representing the dictionary.
        """
        try:
            dist_name = var_dict["distribution"]["name"]
        except KeyError:
            dist_name = var_dict["distribution"]["implements"]
        version = var_dict["distribution"].get("version", "1.0")
        var_type = var_dict["type"]
        unique = var_dict["distribution"]["unique"]
        dist_class = self.find_distribution(dist_name, version=version,
                                            var_type=var_type, unique=unique)
        return dist_class.from_dict(var_dict["distribution"])

    @property
    def distributions(self):
        """All available distributions from fitters, deduplicated."""
        dists = [f.distribution for f in self.fitters]
        # Deduplicate distributions
        new_dists = []
        dist_names = set()
        for dist in dists:
            if dist not in dist_names:
                new_dists.append(dist)
                dist_names.add(dist)
        return new_dists



"""Test the string type distribution inference."""
import pandas as pd
import polars as pl
import pytest
from pytest import mark

from metasyn.distribution.faker import FakerFitter, UniqueFakerDistribution
from metasyn.distribution.freetext import FreeTextFitter
from metasyn.distribution.regex import UniqueRegexDistribution
from metasyn.privacy import BasicPrivacy
from metasyn.var import MetaVar


@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_faker(series_type):
    """Test the faker distribution."""
    dist = FakerFitter(BasicPrivacy()).fit(series_type([1, 2, 3]))
    assert isinstance(dist.to_dict(), dict)
    assert isinstance(dist.draw(), str)
    assert 'faker' in str(dist)
    assert dist.locale == "en_US"
    assert dist.faker_type == "city"
    var = MetaVar("some_city", "string", dist, prop_missing=0.0)
    series_1 = var.draw_series(100, seed=1234)
    series_2 = var.draw_series(100, seed=1234)
    assert all(series_1 == series_2)


@mark.parametrize(
    "series,lang,avg_sentences,avg_words", [
            (pl.Series(
                ["hotdog", "mother", "yes", "tip", "crate",
                 "sink", "dark", "crossbar", "toilet", "grow", "patient"]), "EN", None, 1),
            (pl.Series(["hotdog mother", "yes tip", "crate sink", "dark crossbar", "toilet grow",
                        "patient is"]), "EN", None, 2),
            (pl.Series(["hotdog mother.", "yes tip.", "crate sink.", "dark crossbar.",
                        "toilet grow.", "patient is."]), "EN", 1, 2),
            (pl.Series(["gaat naar school. Ik ben benieuwd.", "Wat is vraag? Wanneer komt hij."]),
             "NL", 2, 6)
    ])
def test_free_text(series, lang, avg_sentences, avg_words):
    """Test whether the free text distribution does the right inference."""
    dist = FreeTextFitter(BasicPrivacy()).fit(series)
    assert dist.locale == lang
    assert dist.avg_sentences == avg_sentences
    assert dist.avg_words == avg_words
    var = MetaVar("some_var", "string", dist, prop_missing=0.0)
    series_1 = var.draw_series(100, seed=1234)
    series_2 = var.draw_series(100, seed=1234)
    assert all(series_1 == series_2)


def test_unique_regex():
    dist = UniqueRegexDistribution(r"[0-9]")
    var = MetaVar("some_var", "string", dist, prop_missing=0.0)

    series = var.draw_series(10, None)
    assert len(series.unique()) == 10

    with pytest.raises(ValueError):
        var.draw()

def test_unique_faker():
    dist = UniqueFakerDistribution("city")
    var = MetaVar("some_var", "string", dist, prop_missing=0.0)

    series = var.draw_series(1000, None)
    assert len(series.unique()) == 1000



"""Module for testing date/time distributions."""
import datetime as dt

import numpy as np
import pandas as pd
import polars as pl
from pytest import mark

from metasyn.distribution.uniform import (
    DateTimeUniformDistribution,
    DateTimeUniformFitter,
    DateUniformDistribution,
    DateUniformFitter,
    TimeUniformDistribution,
    TimeUniformFitter,
)
from metasyn.privacy import BasicPrivacy

all_precision = ["microseconds", "seconds", "minutes", "hours"]
lower = ["10", ""]


@mark.parametrize(
    "lower,upper,precision", [
        ("10:23:45.293852", "16:23:45.293852", "microseconds"),
        ("10:23:45", "16:23:45", "seconds"),
        ("10:23:00", "16:23:00", "minutes"),
        ("10:00:00", "16:00:00", "hours"),
    ]
)
@mark.parametrize("series_type", [pl.Series, pd.Series])
def test_time(lower, upper, precision, series_type):
    """Test for TimeUniformDistribution."""
    lower_iso, upper_iso = dt.time.fromisoformat(lower), dt.time.fromisoformat(upper)
    dist = TimeUniformDistribution(lower, upper, precision)
    all_times = []
    for _ in range(100):
        new_time = dist.draw()
        assert isinstance(new_time, dt.time)
        assert new_time >= lower_iso
        assert new_time <= upper_iso
        for prec in all_precision:
            if prec == precision:
                break
            assert getattr(new_time, prec[:-1]) == 0
        all_times.append(new_time)
    series = series_type(all_times)
    new_dist = TimeUniformFitter(BasicPrivacy()).fit(series)
    assert new_dist.lower >= dist.lower
    assert new_dist.upper <= dist.upper
    assert new_dist.precision == dist.precision


@mark.parametrize("series_type", [pl.Series, pd.Series])
def test_date(series_type):
    """Test for DateUniformDistribution."""
    lower_date, upper_date = "2020-07-09", "2023-08-19"
    lower_iso, upper_iso = dt.date.fromisoformat(lower_date), dt.date.fromisoformat(upper_date)

    dist = DateUniformDistribution(lower_date, upper_date)
    all_dates = []
    for _ in range(100):
        new_date = dist.draw()
        assert isinstance(new_date, dt.date)
        assert new_date >= lower_iso
        assert new_date <= upper_iso
        all_dates.append(new_date)
    series = series_type(all_dates)
    new_dist = DateUniformFitter(BasicPrivacy()).fit(series)
    assert new_dist.lower >= dist.lower
    assert new_dist.upper <= dist.upper
    assert new_dist.precision == dist.precision


@mark.parametrize(
    "lower,upper,precision", [
        ("2020-07-09 10:23:45.293852", "2020-08-09 16:23:45.293852", "microseconds"),
        ("2020-07-09 10:23:45", "2020-08-09 16:23:45", "seconds"),
        ("2020-07-09 10:23:00", "2020-08-09 16:23:00", "minutes"),
        ("2020-07-09 10:00:00", "2020-08-09 16:00:00", "hours"),
        (np.datetime64('2005-02-25T03:30'), np.datetime64('2006-02-25T03:30'), "minutes")
    ]
)
@mark.parametrize("series_type", [pl.Series, pd.Series])
def test_datetime(lower, upper, precision, series_type):
    """Test for DateTimeUniformDistribution."""
    if isinstance(lower, str):
        lower_iso, upper_iso = dt.datetime.fromisoformat(lower), dt.datetime.fromisoformat(upper)
    else:
        lower_iso = lower
        upper_iso = upper
    dist = DateTimeUniformDistribution(lower, upper, precision)
    all_datetimes = []
    for _ in range(100):
        new_time = dist.draw()
        assert isinstance(new_time, dt.datetime)
        assert new_time >= lower_iso
        assert new_time <= upper_iso
        for prec in all_precision:
            if prec == precision:
                break
            assert getattr(new_time, prec[:-1]) == 0
        all_datetimes.append(new_time)

    series = series_type(all_datetimes)
    new_dist = DateTimeUniformFitter(BasicPrivacy()).fit(series)
    assert new_dist.lower >= dist.lower
    assert new_dist.upper <= dist.upper
    assert new_dist.precision == dist.precision



"""Module defining the MetaVar class, which represents a metadata variable."""

from __future__ import annotations

from typing import Any, Dict, Optional, Union

import numpy as np
import polars as pl
from tqdm import tqdm

from metasyn.distribution.base import BaseDistribution
from metasyn.privacy import BasePrivacy, BasicPrivacy
from metasyn.registry import DistributionRegistry
from metasyn.util import get_var_type, set_global_seeds
from metasyn.varspec import DistributionSpec


class MetaVar:
    """Metadata variable describing a column in a MetaFrame.

    MetaVar is a structure that holds all metadata needed to generate a
    synthetic column for it. This is the variable level building block for the
    MetaFrame. It contains the methods to convert a polars Series into a
    variable with an appropriate distribution. The MetaVar class is to the
    MetaFrame what a polars Series is to a DataFrame.

    This class is considered a passthrough class used by the MetaFrame class,
    and is not intended to be used directly by the user.

    Parameters
    ----------
    var_type:
        String containing the variable type, e.g. continuous, string, etc.
    series:
        Series to create the variable from. Series is None by default and in
        this case the value is ignored. If it is not supplied, then the
        variable cannot be fit.
    name:
        Name of the variable/column.
    distribution:
        Distribution to draw random values from. Can also be set by using the
        fit method.
    prop_missing:
        Proportion of the series that are missing/NA.
    dtype:
        Type of the original values, e.g. int64, float, etc. Used for type-casting
        back. The default value is "unknown".
    description:
        User-provided description of the variable.
    creation_method:
        A dictionary that contains information on how the variable was created. If None,
        it will be assumed to have been created by the user.
    """

    def __init__(  # noqa: PLR0913
        self,
        name: str,
        var_type: Optional[str],
        distribution: BaseDistribution,
        dtype: str = "unknown",
        description: Optional[str] = None,
        prop_missing: float = 0.0,
        creation_method: Optional[dict] = None,
    ):
        self.name = name
        if var_type is None:
            var_type = get_var_type(pl.Series([distribution.draw()]))
            distribution.draw_reset()
        self.var_type = var_type
        self.distribution = distribution
        self.dtype = dtype
        self.description = description
        self.prop_missing = prop_missing
        self.creation_method = creation_method
        if creation_method is None:
            self.creation_method = {"created_by": "user"}
        if self.prop_missing < -1e-8 or self.prop_missing > 1 + 1e-8:
            raise ValueError(
                f"Cannot create variable '{self.name}' with proportion missing outside range [0, 1]"
            )
        if self.dtype == "unknown":
            if self.var_type == "categorical":
                self.dtype = "Categorical"
            else:
                self.dtype = str(pl.Series([self.distribution.draw()]).dtype)

    def to_dict(self) -> Dict[str, Any]:
        """Create a dictionary from the variable."""
        if self.distribution is None:
            dist_dict = {}
        else:
            dist_dict = self.distribution.to_dict()
        var_dict = {
            "name": self.name,
            "type": self.var_type,
            "dtype": self.dtype,
            "prop_missing": self.prop_missing,
            "distribution": dist_dict,
            "creation_method": self.creation_method,
        }
        if self.description is not None:
            var_dict["description"] = self.description
        return var_dict

    def __str__(self) -> str:
        """Return an easy to read formatted string for the variable."""
        description = f'Description: "{self.description}"\n' if self.description else ""

        if self.distribution is None:
            distribution_formatted = "No distribution information available"
        else:
            distribution_formatted = "\n".join(
                "\t" + line for line in str(self.distribution).split("\n")
            )

        return (
            f'"{self.name}"\n'
            f"{description}"
            f"- Variable Type: {self.var_type}\n"
            f"- Data Type: {self.dtype}\n"
            f"- Proportion of Missing Values: {self.prop_missing:.4f}\n"
            f"- Distribution:\n{distribution_formatted}\n"
        )

    def __repr__(self) -> str:
        return f"MetaVar <{self.name}, {self.distribution.name}>"

    @classmethod
    def fit(
        cls,  # pylint: disable=too-many-arguments
        series: pl.Series,
        dist_spec: Optional[Union[dict, type, BaseDistribution, DistributionSpec]] = None,
        dist_registry: DistributionRegistry = DistributionRegistry.parse("builtin"),
        privacy: BasePrivacy = BasicPrivacy(),
        prop_missing: Optional[float] = None,
        description: Optional[str] = None,
    ) -> MetaVar:
        """Fit distributions to the data.

        If multiple distributions are available for the current data type,
        use the one that fits the data the best.

        While it has no arguments or return values, it will set the
        distribution attribute to the most suitable distribution.

        Parameters
        ----------
        series:
            Data series to fit a distribution to.
        dist_spec:
            The distribution to fit. In case of a string, search for it
            using the aliases of all distributions. Otherwise use the
            supplied distribution (class). Examples of allowed strings are:
            "normal", "uniform", "faker.city.nl_NL". If not supplied, fit
            the best available distribution for the variable type.
        dist_registry:
            Distribution registry that is used for fitting.
        privacy:
            Privacy level to use for fitting the series.
        prop_missing:
            Proportion of the values missing, default None.
        description:
            Description for the variable.
        """
        if not isinstance(series, pl.Series):
            series = pl.Series(series)
        var_type = get_var_type(series)
        dist_spec = DistributionSpec.parse(dist_spec)
        distribution, fitter = dist_registry.fit(series, var_type, dist_spec, privacy)
        if prop_missing is None:
            prop_missing = (len(series) - len(series.drop_nulls())) / len(series)
        return cls(
            series.name,
            var_type,
            distribution=distribution,
            dtype=str(series.dtype),
            description=description,
            prop_missing=prop_missing,
            creation_method=dist_spec.get_creation_method(fitter),
        )

    def draw(self) -> Any:
        """Draw a random item for the variable in whatever type is required."""
        # Return NA's -> None
        if self.prop_missing is not None and np.random.rand() < self.prop_missing:
            return None
        return self.distribution.draw()

    def draw_series(self, n: int, seed: Optional[int], progress_bar: bool = True) -> pl.Series:
        """Draw a new synthetic series from the metadata.

        Parameters
        ----------
        n:
            Length of the series to be created.
        seed:
            Seed value for the internal random number generator. Set this to ensure reproducibility.
        progress_bar:
            Whether to display a progress bar.

        Returns
        -------
        polars.Series:
            Polars series with the synthetic data.
        """
        if seed is not None:
            set_global_seeds(seed)

        self.distribution.draw_reset()

        is_not_na = np.random.rand(n) >= self.prop_missing
        n_draw: int = np.sum(is_not_na)  # type: ignore
        try:
            not_na_values = self.distribution.draw_list(n_draw)
        except NotImplementedError:
            not_na_values = [self.distribution.draw()
                             for _ in tqdm(range(n_draw), disable=not progress_bar, leave=False,
                                           desc="synthesizing")]

        # Mix the values with Nones
        cum_not_na = np.cumsum(is_not_na)
        value_list = [not_na_values[cum_not_na[i]-1] if is_not_na[i] else None for i in range(n)]
        pl_type = self.dtype.split("(")[0]

        # Workaround for polars issue with numpy 2.0
        if pl_type == "Boolean":
            value_list = [None if x is None else bool(x) for x in value_list]

        # Some dtypes have extra information, discard that
        return pl.Series(value_list, dtype=getattr(pl, pl_type))

    @classmethod
    def from_dict(
        cls,
        var_dict: Dict[str, Any],
        plugins: Union[None, str, list[str]] = None,
    ) -> MetaVar:
        """Restore variable from dictionary.

        Parameters
        ----------
        plugins:
            Plugins to use to create the variable. If None,
            use all installed/available plugins.
        var_dict:
            This dictionary contains all the variable and distribution
            information to recreate it from scratch.

        Returns
        -------
        MetaVar:
            Initialized metadata variable.
        """
        dist_registry = DistributionRegistry.parse(plugins)
        dist = dist_registry.from_dict(var_dict)
        return cls(
            name=var_dict["name"],
            var_type=var_dict["type"],
            distribution=dist,
            prop_missing=var_dict["prop_missing"],
            dtype=var_dict["dtype"],
            description=var_dict.get("description", None),
            creation_method=var_dict.get("creation_method", {"created_by": "unknown"}),
        )



"""Test whether the privacy functionality works."""
import pytest

from metasyn.distribution.base import metafit
from metasyn.distribution.categorical import MultinoulliFitter
from metasyn.privacy import BasicPrivacy, get_privacy


@metafit(privacy_type="test")
class OtherMultinoulli(MultinoulliFitter):
    """Test privacy version of multinoulli."""


def test_base_privacy():
    """Test whether the builtin version of privacy (no extra) works as expected."""
    privacy = BasicPrivacy()
    priv_dict = privacy.to_dict()
    assert isinstance(priv_dict, dict)
    assert priv_dict["name"] == "none"
    assert privacy.name == "none"
    assert isinstance(priv_dict["parameters"], dict)
    assert len(priv_dict["parameters"]) == 0
    assert privacy.is_compatible(MultinoulliFitter)
    assert not privacy.is_compatible(OtherMultinoulli)


def test_import_error():
    """Test whether trying to use a non-existing privacy type raises an error."""
    with pytest.raises(ImportError):
        get_privacy("nonexistentprivacyplugin")



"""Module for distribution and variable specifications."""
from __future__ import annotations

# from metasyn.util import VarSpec
from dataclasses import dataclass, field
from typing import Any, Optional, Union

from metasyn.distribution.base import BaseDistribution, BaseFitter
from metasyn.privacy import BasePrivacy, get_privacy
from metasyn.util import ALL_VAR_TYPES


@dataclass
class DistributionSpec():
    """Specification that determines which distribution is selected.

    It has the following attributes:
    - name: Which distribution is chosen.
    - unique: Whether the distribution should be unique.
    - parameters: The parameters of the distribution as defined by name.
    - fit_kwargs: Fitting keyword arguments to be used while fitting the distribution.
    - version: Version of the distribution to fit.
    """

    name: Optional[str] = None
    unique: Optional[bool] = None
    parameters: Optional[dict] = None
    fit_kwargs: dict = field(default_factory=dict)
    version: Optional[str] = None
    distribution: Optional[BaseDistribution] = None

    def __post_init__(self):
        if self.name is None:
            if self.version is not None:
                raise ValueError("Cannot create DistributionSpec with attribute 'version' but "
                                 "without attribute 'name'.")
            if self.parameters is not None:
                raise ValueError("Cannot create DistributionSpec with attribute 'parameters' but "
                                 "without attribute 'name'.")
            if len(self.fit_kwargs) > 0:
                raise ValueError("Cannot create DistributionSpec with attribute 'fit_kwargs' that"
                                 " is not empty but without attribute 'name'.")


    @classmethod
    def parse(cls, dist_spec: Optional[Union[dict, type[BaseDistribution], BaseDistribution,
                                             DistributionSpec, str]],
              unique: Optional[bool] = None,
              ) -> DistributionSpec:
        """Create a DistributionSpec instance from a variety of inputs.

        Parameters
        ----------
        dist_spec:
            Specification for the distribution in several types.
        unique:
            Whether the distribution is unique. This is only taken into account
            if dist_spec is None or a string.

        Returns
        -------
            A instantiated version of the dist_spec that has the DistributionSpec type.

        Raises
        ------
        TypeError
            If the input has the wrong type and cannot be parsed.
        """
        if isinstance(dist_spec, BaseDistribution):
            dist_dict = {key: value for key, value in dist_spec.to_dict().items()
                         if key in ["name", "version", "unique", "parameters"]}
            return cls(**dist_dict, distribution=dist_spec)
        if isinstance(dist_spec, str):
            return cls(name=dist_spec, unique=unique)
        if dist_spec is None:
            return cls(unique=unique)
        if isinstance(dist_spec, dict):
            return cls(**dist_spec)
        if isinstance(dist_spec, DistributionSpec):
            return dist_spec
        if issubclass(dist_spec, BaseDistribution):
            return cls(name=dist_spec.name, unique=dist_spec.unique)
        raise TypeError("Error parsing distribution specification of unknown type "
                        f"'{type(dist_spec)}' with value '{dist_spec}'")

    @property
    def fully_specified(self) -> bool:
        """Indicate whether the distribution is suitable for datafree creation.

        Returns
        -------
            A flag that indicates whether a distribution can be generated from the values
            that are specified (not None).
        """
        return self.name is not None and self.parameters is not None

    def get_creation_method(self, fitter: Optional[BaseFitter]) -> dict:
        """Create a dictionary on how the distribution was created.

        Parameters
        ----------
        privacy
            Privacy object with which the dictionary is being created.

        Returns
        -------
            Dictionary containing all the non-default settings for the creation method.
        """
        ret_dict: dict[str, Any] = {"created_by": "metasyn"}
        dist_dict = {var: getattr(self, var) for var in ["name", "unique", "parameters", "version"]
                     if getattr(self, var) is not None}
        if len(dist_dict) != 0:
            ret_dict["distribution"] = dist_dict
        fit_dict = {}
        if fitter is not None:
            fit_dict = fitter.to_dict()
        if len(self.fit_kwargs) > 0:
            fit_dict["fit_kwargs"] = self.fit_kwargs

        if len(fit_dict) != 0:
            ret_dict["fitter"] = fit_dict

        return ret_dict

    def __str__(self):
        return f"DistSpec<{self.name}, {self.unique}, {self.parameters}, {self.distribution}>"


class VarSpec():  # pylint: disable=too-few-public-methods
    """Data class for storing the specifications for variables.

    Parameters
    ----------
    name:
        Name of the variable/column.
    distribution:
        Distribution to use for fitting/finding the distribution.
        Leave at None to allow metasyn to find the most suitable distribution
        automatically.

        >>> # Use normal distribution
        >>> distribution="normal"
        >>> # Use normal distribution with mean 0, standard deviation 1
        >>> distribution=NormalDistribution(0, 1)

    unique:
        To set a column to be unique/key.
        This is only available for the integer and string datatypes. Setting a variable
        to unique ensures that the synthetic values generated for this variable are unique.
        This is useful for ID or primary key variables, for example. The parameter...
        is ignored when the distribution is set manually. For example:
        {"unique": True}, which sets the variable to be unique or {"unique": False} which
        forces the variable to be not unique. If the uniqueness is not specified, it is
        assumed to be not unique, but gives a warning if metasyn thinks it should be.
    privacy:
        Set the privacy level for a variable, e.g.: DifferentialPrivacy(epsilon=10).
    prop_missing:
        Proportion of missing values for a variable.
    description:
        Set the description of a variable.
    data_free:
        Whether this variable/column is to be generated from scratch or from an existing column
        in the dataframe.
    var_type:
        Manually set the variable type of the columns (used mainly for data_free columns).
    """

    def __init__(
            self,
            name: str,
            distribution: Optional[Union[dict, type[BaseDistribution], BaseDistribution,
                                                DistributionSpec, str]] = None,
            unique=None,
            privacy: Optional[BasePrivacy] = None,
            prop_missing: Optional[float] = None,
            description: Optional[str] = None,
            data_free: Optional[bool] = None,
            var_type: Optional[str] = None):

        self.name = name
        self.dist_spec = DistributionSpec.parse(distribution, unique)
        self.privacy = privacy
        self.prop_missing = prop_missing
        self.description = description
        self.data_free = data_free
        self.var_type = var_type
        self.__post_init__()

    def __post_init__(self):
        # Convert the the privacy attribute if it is a dictionary.
        if isinstance(self.privacy, dict):
            self.privacy = get_privacy(**self.privacy)
        if self.data_free and not self.dist_spec.fully_specified:
            raise ValueError("Error creating variable specification: data free variable should have"
                            f" 'name' and 'parameters'. {self}")
        if self.var_type is not None and self.var_type not in ALL_VAR_TYPES:
            raise ValueError(f"Cannot create variable '{self.name}': unknown variable type "
                             f"'{self.var_type}'. Choose from {ALL_VAR_TYPES}.")

    @classmethod
    def from_dict(cls, var_dict: dict) -> VarSpec:
        """Create a variable specification from a dictionary.

        Parameters
        ----------
        var_dict
            Dictionary to parse the specification from.

        Returns
        -------
            A VarSpec instance.
        """
        return cls(**var_dict)


@dataclass
class VarDefaults():
    """Dataclass for variable defaults.

    Parameters
    ----------
    data_free:
        Whether the variable is completely synthetic or is based on real data.
    prop_missing:
        Proportion of missing values.
    distribution:
        Dictionary containing default distributions for each variable type.
    privacy:
        Privacy to be used by default for estimating distributions.
    """

    data_free: bool = False
    prop_missing: Optional[float] = None
    distribution: Optional[dict] = None
    privacy: Optional[BasePrivacy] = None

    def __post_init__(self):
        if isinstance(self.privacy, dict):
            self.privacy = get_privacy(**self.privacy)
        elif not isinstance(self.privacy, BasePrivacy):
            self.privacy = get_privacy("none")




"""Module implementing NA distributions.

This module contains a single class for creating distributions that only
return NA.
"""


from metasyn.distribution.base import BaseDistribution, BaseFitter, builtin_fitter, metadist


@metadist(name="core.na", var_type=["continuous", "discrete", "categorical", "string"])
class NADistribution(BaseDistribution):
    """Distribution that always returns NA values (None)."""

    @classmethod
    def default_distribution(cls, var_type=None) -> BaseDistribution: # noqa: ARG003
        return cls()

    def draw(self):
        return None

    def _param_dict(self):
        return {}

    @classmethod
    def _param_schema(cls):
        return {}

    def information_criterion(self, values): # noqa: ARG002
        return 1e10

@builtin_fitter(distribution=NADistribution,
                var_type=["continuous", "discrete", "categorical", "string"])
class NAFitter(BaseFitter):
    """Fitter for NA distribution."""

    def _fit(self, series):  # noqa: ARG002
        return self.distribution()
